import time
import unittest

from MiscUtils.Funcs import (
    asclocaltime, commas, charWrap, wordWrap, excstr,
    hostName, localIP, localTimeDelta, positiveId,
    safeDescription, timestamp, uniqueId, valueForString)


class TestFuncs(unittest.TestCase):
    """Unit tests for the functions in MiscUtils.Funcs."""

    def testCommas(self):
        testSpec = '''
            0 '0'
            0.0 '0.0'
            1 '1'
            11 '11'
            111 '111'
            1111 '1,111'
            11111 '11,111'
            1.0 '1.0'
            11.0 '11.0'
            1.15 '1.15'
            12345.127 '12,345.127'
            -1 '-1'
            -11 '-11'
            -111 '-111'
            -1111 '-1,111'
            -11111 '-11,111'
        '''
        tests = testSpec.split()
        count = len(tests)
        i = 0
        while i < count:
            source = eval(tests[i])
            result = eval(tests[i+1])
            self.assertEqual(commas(source), result)
            # also try the source as a string instead of a number
            source = eval(f"'{tests[i]}'")
            self.assertEqual(commas(source), result)
            i += 2

    def testCharWrap(self):
        self.assertEqual(charWrap("""
            Sparse is better than dense.
            Readability counts.""", 34, 16), """
            Sparse is better than
                dense.
            Readability counts.""")

    def testWordWrap(self):
        # an example with some spaces and newlines
        msg = '''Arthur:  "The Lady of the Lake, her arm clad in the purest \
shimmering samite, held aloft Excalibur from the bosom of the water, \
signifying by Divine Providence that I, Arthur, was to carry \
Excalibur. That is why I am your king!"

Dennis:  "Listen. Strange women lying in ponds distributing swords is \
no basis for a system of government. Supreme executive power derives \
from a mandate from the masses, not from some farcical aquatic ceremony!"'''

        for margin in range(20, 200, 29):
            if margin == 78:
                s = wordWrap(msg)
            else:
                s = wordWrap(msg, margin)
            for line in s.splitlines():
                self.assertLessEqual(
                    len(line), margin,
                    f'len={len(line)}, margin={margin}, line={line!r}')
            self.assertEqual(msg.split(), s.split())

    def testExcstr(self):
        self.assertEqual(excstr(None), None)
        self.assertEqual(
            excstr(ValueError('Kawoom!')), 'ValueError: Kawoom!')

    def testHostName(self):
        # About all we can do is invoke hostName() to see that no exceptions
        # are thrown, and do a little type checking on the return type.
        host = hostName()
        self.assertIsNotNone(host)
        self.assertIsInstance(host, str)
        self.assertEqual(host, host.lower())

    def testLocalIP(self):
        ip = localIP()
        self.assertTrue(ip)
        self.assertFalse(ip.startswith('127.'))
        self.assertEqual(localIP(), ip)  # second invocation
        self.assertEqual(localIP(useCache=None), ip)
        self.assertEqual(
            localIP(remote=None, useCache=None), ip,
            'See if this works: localIP(remote=None).'
            ' If this fails, dont worry.')
        self.assertEqual(
            localIP(remote=('www.hostname.and.domain.are.invalid', 80),
                    useCache=None), ip)

    def testPositiveId(self):
        # About all we can do is invoke positiveId()
        # to see that no exceptions are thrown and the result is positive.
        self.assertIsInstance(positiveId(self), int)
        self.assertGreater(positiveId(self), 0)

    def testSafeDescription(self):
        desc = safeDescription

        # basics:
        s = desc(1).replace('type=', 'class=')
        self.assertEqual(s, "what=1 class=<class 'int'>")
        s = desc(1, 'x').replace('type=', 'class=')
        self.assertEqual(s, "x=1 class=<class 'int'>")
        s = desc('x').replace('type=', 'class=')
        s = s.replace("<type 'string'>", "<class 'str'>")
        self.assertEqual(s, "what='x' class=<class 'str'>")

        class Dummy:
            pass

        s = desc(Dummy())
        self.assertIn('Dummy object', s, repr(s))

        # okay now test that safeDescription eats exceptions from repr():
        class Bogus:
            def __repr__(self):
                raise KeyError('bogus')

        b = Bogus()
        try:
            s = desc(b)
        except Exception:
            s = 'failure: should not throw exception'

        self.assertIn("(exception from repr(obj): KeyError: 'bogus')", s)

    def testAsclocaltime(self):
        self.assertEqual(len(asclocaltime()), 24)
        t = time.time()
        self.assertEqual(asclocaltime(t), time.asctime(time.localtime(t)))

    def testTimestamp(self):
        d = timestamp()
        self.assertIsInstance(d, dict)
        self.assertEqual(','.join(sorted(d)), 'condensed,dashed,pretty,tuple')
        self.assertEqual(len(d['tuple']), 6)
        self.assertEqual(len(d['condensed']), 14)
        self.assertEqual(len(d['pretty']), 19)
        self.assertEqual(len(d['dashed']), 19)
        t = time.time()
        d = timestamp(t)
        t = time.localtime(t)[:6]
        self.assertEqual(d['tuple'], t)
        self.assertEqual(
            d['condensed'], '{:4d}{:02d}{:02d}{:02d}{:02d}{:02d}'.format(*t))
        self.assertEqual(
            d['condensed'],
            d['pretty'].replace('-', '').replace(':', '').replace(' ', ''))
        self.assertEqual(d['condensed'], d['dashed'].replace('-', ''))

    def testLocalTimeDelta(self):
        d = localTimeDelta()
        self.assertEqual(d.microseconds, 0)
        self.assertEqual(d.seconds % 3600, 0)
        self.assertTrue(-1 <= d.days < 1)
        d = localTimeDelta(time.time())
        self.assertEqual(d.microseconds, 0)
        self.assertEqual(d.seconds % 3600, 0)
        self.assertTrue(-1 <= d.days < 1)

    def testUniqueId(self):
        past = set()

        def checkId(i):
            self.assertIsInstance(i, str, type(i))
            self.assertEqual(len(i), 32)
            for c in i:
                self.assertTrue(c in '0123456789abcdef')
            self.assertFalse(i in past)
            past.add(i)

        for n in range(10):
            checkId(uniqueId())
            checkId(uniqueId(None))
            checkId(uniqueId(n))
            checkId(uniqueId(forObject=checkId))

    def testValueForString(self):
        evalCases = '''
            1
            9223372036854775808
            5.5
            True
            False
            None
            [1]
            ['a']
            {'x':1}
            (1, 2, 3)
            'a'
            "z"
            """1234"""
        '''

        stringCases = '''
            kjasdfkasdf
            2389234lkdsflkjsdf
            *09809
        '''

        evalCases = [s.strip() for s in evalCases.strip().splitlines()]
        for case in evalCases:
            value = valueForString(case)
            evalCase = eval(case)
            self.assertEqual(
                value, evalCase,
                f'case={case!r}, valueForString()={value!r},'
                f' eval()={evalCase!r}')

        stringCases = [s.strip() for s in stringCases.strip().splitlines()]
        for case in stringCases:
            value = valueForString(case)
            self.assertEqual(
                value, case,
                f'case={case!r}, valueForString()={value!r}')



import os
from Examples.ExamplePage import ExamplePage


class PSPExamplePage(ExamplePage):

    def cornerTitle(self):
        return "PSP Examples"

    def writeOtherMenu(self):
        self.menuHeading('Other')
        self.menuItem(
            f'View source of<br/>{self.title()}',
            self.request().uriWebwareRoot() + 'PSP/Examples/View?filename=' +
            os.path.basename(self.request().serverSidePath()))



"""WebUtils.Funcs

This module provides some basic functions that are useful
in HTML and web development.

You can safely import * from WebUtils.Funcs if you like.
"""

# Aliases for URL encoding and decoding functions:
from urllib.parse import quote_plus as urlEncode, unquote_plus as urlDecode

__all__ = [
    'htmlEncode', 'htmlEncodeStr', 'htmlDecode', 'urlEncode', 'urlDecode',
    'htmlForDict', 'requestURI', 'normURL']

htmlForNone = '-'  # used by htmlEncode

htmlCodes = (
    ('&', '&amp;'),
    ('<', '&lt;'),
    ('>', '&gt;'),
    ('"', '&quot;'),
)

htmlCodesReversed = tuple(reversed(htmlCodes))


def htmlEncode(what, codes=htmlCodes):
    """Return the HTML encoded version of the given object.

    The optional 'codes' parameter allows passing custom translations.
    """
    if what is None:
        return htmlForNone
    if hasattr(what, 'html'):
        # allow objects to specify their own translation to html
        # via a method, property or attribute
        html = what.html
        if callable(html):
            html = html()
        return html
    what = str(what)
    return htmlEncodeStr(what, codes)


def htmlEncodeStr(s, codes=htmlCodes):
    """Return the HTML encoded version of the given string.

    This is useful to display a plain ASCII text string on a web page.

    The optional 'codes' parameter allows passing custom translations.
    """
    for c, e in codes:
        s = s.replace(c, e)
    return s


def htmlDecode(s, codes=htmlCodesReversed):
    """Return the ASCII decoded version of the given HTML string.

    This does NOT remove normal HTML tags like <p>.
    It is the inverse of htmlEncode().

    The optional 'codes' parameter allows passing custom translations.
    """
    for c, e in codes:
        s = s.replace(e, c)
    return s


def htmlForDict(d, addSpace=None, filterValueCallBack=None,
                maxValueLength=None, topHeading=None, isEncoded=None):
    """Return HTML string with a table where each row is a key/value pair."""
    if not d:
        return ''
    html = ['<table class="NiceTable">\n']
    append = html.append
    if topHeading:
        html.append('<tr class="TopHeading">')
        if isinstance(topHeading, tuple):
            append('<th>{}</th><th>{}</th>'.format(*topHeading))
        else:
            append('<th colspan="2">{}</th>'.format(topHeading))
        append('</tr>\n')
    for key in sorted(d):
        value = d[key]
        if addSpace and key in addSpace:
            target = addSpace[key]
            value = (target + ' ').join(value.split(target))
        if filterValueCallBack:
            value = filterValueCallBack(value, key, d)
        if maxValueLength and not isEncoded:
            value = str(value)
            if len(value) > maxValueLength:
                value = value[:maxValueLength - 3] + '...'
        key = htmlEncode(key)
        if not isEncoded:
            value = htmlEncode(value)
        append(f'<tr><th>{key}</th><td>{value}</td></tr>\n')
    append('</table>')
    return ''.join(html)


def requestURI(env):
    """Return the request URI for a given CGI-style dictionary.

    Uses REQUEST_URI if available, otherwise constructs and returns it
    from SCRIPT_URL, SCRIPT_NAME, PATH_INFO and QUERY_STRING.
    """
    uri = env.get('REQUEST_URI')
    if uri is None:
        uri = env.get('SCRIPT_URL')
        if uri is None:
            uri = env.get('SCRIPT_NAME', '') + env.get('PATH_INFO', '')
        query = env.get('QUERY_STRING', '')
        if query != '':
            uri += '?' + query
    return uri


def normURL(path):
    """Normalizes a URL path, like os.path.normpath.

    Acts on a URL independent of operating system environment.
    """
    if not path:
        return
    startSlash = path.startswith('/')
    endSlash = path.endswith('/')
    comps = path.split('/')
    newComps = []
    for comp in comps:
        if comp in ('', '.'):
            continue
        if comp != '..':
            newComps.append(comp)
        elif newComps:
            newComps.pop()
    path = '/'.join(newComps)
    if path and endSlash:
        path += '/'
    if startSlash:
        path = '/' + path
    return path



"""M2Crypto-enhanced transport for PickleRPC

This lets you use M2Crypto for SSL encryption.

Based on m2xmlrpclib.py which is
Copyright (c) 1999-2002 Ng Pheng Siong. All rights reserved.
"""

from io import BytesIO

from M2Crypto import SSL, httpslib, m2urllib  # pylint: disable=import-error

from .PickleRPC import Transport

__version__ = 1  # version of M2PickleRPC


class M2Transport(Transport):

    user_agent = f"M2PickleRPC.py/{__version__} - {Transport.user_agent}"

    def __init__(self, ssl_context=None):
        if ssl_context is None:
            self.ssl_ctx = SSL.Context('sslv23')
        else:
            self.ssl_ctx = ssl_context

    def make_connection(self, host, port=None):
        if port is None:
            host, port = m2urllib.splitport(host)
        return httpslib.HTTPS(host, int(port), ssl_context=self.ssl_ctx)

    # Workarounds below are necessary because M2Crypto seems to
    # return from fileobject.read() early!  So we have to call it
    # over and over to get the full data.

    def parse_response(self, f):
        """Workaround M2Crypto issue mentioned above."""
        sio = BytesIO()
        while True:
            chunk = f.read()
            if not chunk:
                break
            sio.write(chunk)
        sio.seek(0)
        return Transport.parse_response(self, sio)

    def parse_response_gzip(self, f):
        """Workaround M2Crypto issue mentioned above."""
        sio = BytesIO()
        while True:
            chunk = f.read()
            if not chunk:
                break
            sio.write(chunk)
        sio.seek(0)
        return Transport.parse_response_gzip(self, sio)



#!/usr/bin/env python3

"""Helper script for the feature provided by the IncludeEditLink setting."""

import os
import sys
from email import message_from_file
from subprocess import run

editor = 'Vim'  # your favorite editor

defaultArgs = [editor, ' +{line}', '{filename}']

# add an entry for your favorite editor here if it does not already exist
editorArgs = {
    'Emacs':
        ['gnuclient', '+{line}',  '{filename}'],
    'Geany':
        ['geany',  '-l', '{line}', '{filename}'],
    'Geany (Windows)':
        ['C:/Program Files (x86)/Geany/bin/geany.exe',
         '-l', '{line}', '{filename}'],
    'gedit':
        ['gedit', '+{line}', '{filename}'],
    'jEdit':
        ['jedit', '{filename}', '+line:{line}'],
    'jEdit (Windows)':
        ['C:/Program Files/jEdit/jedit.exe', '{filename}', '+line:{line}'],
    'Kate':
        ['kate', '-u', '-l', '{line}', '{filename}'],
    'Komodo':
        ['komodo', '-l', '{line}', '{filename}'],
    'KWrite':
        ['kwrite', '--line', '{line}', '{filename}'],
    'Notepad++ (Windows)':
        ['C:/Program Files/Notepad++/notepad++.exe', '-n{line}', '{filename}'],
    'PSPad (Windows)':
        ['C:/Program Files (x86)/PSPad editor/PSPad.exe',
         '-{line}', '{filename}'],
    'SciTE':
        ['scite', '{filename}', '-goto:{line}'],
    'SciTE (Windows)':
        ['C:/wscite/SciTE.exe', '{filename}', '-goto:{line}'],
    'Sublime Text (Windows)':
        ['C:/Program Files/Sublime Text 3/subl.exe', '{filename}:{line}'],
    'Vim':
        ['gvim', '+{line}', '{filename}'],
}


def transform(params):
    """Transform EditFile parameters.

    As an example, if you are under Windows and your edit file
    has a Unix filename, then it is transformed to a UNC path.
    """
    filename = params['filename']
    if os.sep == '\\' and filename.startswith('/'):
        filename = os.path.normpath(filename[1:])
        hostname = params['hostname'].split(':', 1)[0]
        smbPath = fr'\\{hostname}\root'
        filename = os.path.join(smbPath, filename)
        params['filename'] = filename
    return params


def openFile(params):
    """Open editor with file specified in parameters."""
    params = {key.lower(): value for key, value in params.items()}
    params = transform(params)
    args = editorArgs.get(editor, defaultArgs)
    args[1:] = [arg.format(**params) for arg in args[1:]]
    print(' '.join(args))
    run(args)


def parseFile(filename):
    """Parse the Webware EditFile."""
    openFile(message_from_file(open(filename)))


if __name__ == '__main__':
    parseFile(sys.argv[1])



"""ExpansiveHTMLForException.py

Create expansive HTML for exceptions using the CGITraceback module.
"""

from WebUtils import CGITraceback

HTMLForExceptionOptions = {
    'table': 'background-color:#F0F0F0;font-size:10pt',
    'default': 'color:#000',
    'row.location': 'color:#009',
    'row.code': 'color:#900',
    'editlink': None,
}


def expansiveHTMLForException(context=5, options=None):
    """Create expansive HTML for exceptions."""
    if options:
        opt = HTMLForExceptionOptions.copy()
        opt.update(options)
    else:
        opt = HTMLForExceptionOptions
    return CGITraceback.html(context=context, options=opt)


# old (deprecated) alias
ExpansiveHTMLForException = expansiveHTMLForException



"""Test Webware Testing context"""

import time
import unittest

from .AppTest import AppTest


class TestTesting(AppTest, unittest.TestCase):

    settings = dict(
        PrintConfigAtStartUp=False,
        ExtraPathInfo=False
    )

    def testStartPage(self):
        r = self.testApp.get('/').click('Testing')
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Testing/')
        self.assertEqual(r.content_type, 'text/html')
        self.assertTrue(r.text.startswith('<!DOCTYPE html>'))
        r.mustcontain(
            '<title>Testing</title>',
            '<h2 style="text-align:center">Test cases</h2>',
            '<td>1.</td>', '<td>2.</td>', '<td>3.</td>', '<td>17.</td>')
        headings = [h.string for h in r.html.find('table').find('tr').children]
        self.assertEqual(headings, ['#', 'URL', 'Expectation'])

    def getTestCaseUrls(self, n, expect):
        r = self.testApp.get('/Testing/')
        case = list(r.html.find('table').find_all('tr')[n].children)
        self.assertEqual(case[0].string, f'{n}.')
        urls = [a['href'] for a in case[1].find_all('a')]
        self.assertGreater(len(urls), 0, 'Expected at least one URL')
        s = ' '.join(str(s) for s in case[2].contents)
        s = s.replace('<br/>', ' ')
        s = ' '.join(s.split())
        self.assertIn(expect, s, 'Unexpected test')
        return urls

    def getTestCaseUrl(self, n, expect):
        urls = self.getTestCaseUrls(n, expect)
        self.assertEqual(len(urls), 1, 'Expected only one URL')
        return urls[0]

    def testCase1(self):
        urls = self.getTestCaseUrls(1, 'Show /Examples/Welcome')
        for url in urls:
            r = self.testApp.get(url)
            self.assertEqual(r.status, '200 OK')
            r.mustcontain(
                '<title>Welcome</title>', 'Welcome to Webware for Python')

    def testCase2(self):
        urls = self.getTestCaseUrls(2, 'Show admin pages')
        for url in urls:
            r = self.testApp.get(url)
            self.assertEqual(r.status, '200 OK')
            r.mustcontain(
                '<title>LoginPage</title>',
                '<div id="CornerTitle">Webware Admin</div>',
                'Logins to admin pages are disabled')

    def testCase3(self):
        url = self.getTestCaseUrl(3, 'Redirect to /Admin/')
        r = self.testApp.get(url)
        self.assertEqual(r.status, '301 Moved Permanently')
        r = r.follow()
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Admin/')
        r.mustcontain('<div id="CornerTitle">Webware Admin</div>')

    def testCase4(self):
        url = self.getTestCaseUrl(4, 'Redirect to /Examples/')
        r = self.testApp.get(url)
        self.assertEqual(r.status, '301 Moved Permanently')
        r = r.follow()
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.request.path, '/Examples/')
        r.mustcontain('<div id="CornerTitle">Webware Examples</div>')

    def testCase5(self):
        urls = self.getTestCaseUrls(5, 'Display the file')
        for url in urls:
            r = self.testApp.get(url)
            self.assertEqual(r.status, '200 OK')
            r.mustcontain('This is <code style="color:#558">File.html</code>'
                          ' in <code style="color:#555">/Testing/Dir/</code>')

    def testCase6(self):
        urls = self.getTestCaseUrls(6, 'Display the index file')
        for url in urls:
            r = self.testApp.get(url)
            if r.status_int == 301:
                r = r.follow()
            self.assertEqual(r.status, '200 OK')
            self.assertEqual(r.request.path, '/Testing/Dir/')
            r.mustcontain('This is <code style="color:#558">index.html</code>'
                          ' in <code style="color:#555">/Testing/Dir/</code>')

    def testCase7(self):
        urls = self.getTestCaseUrls(7, 'Not Found if ExtraPathInfo is not set')
        for url in urls:
            if self.app.setting('ExtraPathInfo'):
                r = self.testApp.get(url)
                self.assertEqual(r.status, '200 OK')
                r.mustcontain(
                    '<title>Welcome</title>', 'Welcome to Webware for Python',
                    no='extraURLPath')
            else:
                r = self.testApp.get(url, expect_errors=True)
                self.assertEqual(r.status, '404 Not Found')
                r.mustcontain(
                    '<h1 style="color:#008">Error 404</h1>',
                    'The page you requested,'
                    f' <strong>{url}</strong>, was not found')

    def testCase8(self):
        urls = self.getTestCaseUrls(8, 'Not Found or if ExtraPathInfo is set')
        for url in urls:
            if self.app.setting('ExtraPathInfo'):
                r = self.testApp.get(url)
                self.assertEqual(r.status, '200 OK')
                if url.startswith('/Examples/'):
                    url = url[9:]
                r.mustcontain(
                    '<title>Welcome</title>', 'Welcome to Webware for Python',
                    f'<li>extraURLPath: <code>{url}</code></li>')
            else:
                r = self.testApp.get(url, expect_errors=True)
                self.assertEqual(r.status, '404 Not Found')

    def testCase9(self):
        urls = self.getTestCaseUrls(9, 'test servlet displays extra path info')
        for url in urls:
            if self.app.setting('ExtraPathInfo'):
                r = self.testApp.get(url)
                self.assertEqual(r.status, '200 OK')
                extraPathInfo = url.partition('/Testing/Servlet')[2]
                extraPathInfo = extraPathInfo.partition('?')[0]
                r.mustcontain(
                    '<title>Test of extra path info.</title>',
                    '<h2>Webware Testing Servlet</h2>',
                    '<h3>Test of extra path info.</h3>',
                    f'<p>extraURLPath = <code>{extraPathInfo}</code></p>')
            else:
                r = self.testApp.get(url, expect_errors=True)
                self.assertEqual(r.status, '404 Not Found')

    def testCase10(self):
        url = self.getTestCaseUrl(10, 'IncludeURLTest test')
        r = self.testApp.get(url)
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>IncludeURLTest</title>',
            '<h2>IncludeURLTest</h2>', '<h2>IncludeURLTest2</h2>',
            'Including the Main servlet of the Testing context:',
            '<h2 style="text-align:center">Test cases</h2>',
            '<h4>IncludeURLTest complete.</h4>')

    def testCase11(self):
        url = self.getTestCaseUrl(11, 'lower level IncludeURLTest test')
        r = self.testApp.get(url)
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>IncludeURLTest2</title>',
            '<h2>IncludeURLTest2</h2>',
            'This is the second part of the URL test code.',
            no=['<h2>IncludeURLTest</h2>', 'Test cases'])

    def testCase12(self):
        url = self.getTestCaseUrl(12, 'Forward1Target')
        r = self.testApp.get(url)
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Forward1Target</title>',
            '<h2><code>Forward1Target</code></h2>',
            no='Test cases')

    def testCase13(self):
        url = self.getTestCaseUrl(13, 'Dir/Forward2Target')
        r = self.testApp.get(url)
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Forward2Target</title>',
            '<h2><code>Forward2Target</code></h2>',
            no='Test cases')

    def testCase14(self):
        url = self.getTestCaseUrl(14, 'Forward3Target')
        r = self.testApp.get(url)
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Forward3Target</title>',
            '<h2><code>Forward3Target</code></h2>',
            no='Test cases')

    def testCase15(self):
        url = self.getTestCaseUrl(15, 'Test of WebUtils.FieldStorage')
        r = self.testApp.get(url)
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>FieldStorage</title>',
            'FieldStorage class can be tested here', 'press the button',
            no=['Test cases', 'works as expected'])
        r = r.form.submit('testbutton')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>FieldStorage</title>',
            'Everything ok', 'FieldStorage works as expected',
            no=['Test cases', 'press the button'])
        r = r.click('Back to the test cases overview.')
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>Testing</title>', 'Test cases')

    def testCase16(self):
        url = self.getTestCaseUrl(16, 'Test of HTTPResponse.setCookie')
        r = self.testApp.get(url)
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>SetCookie</title>', 'The time right now is:',
            '<h2>Cookies being sent:</h2>', "'onclose' sends:",
            no="'onclose' =")
        r = r.goto(url)
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>SetCookie</title>', 'The time right now is:',
            '<h2>Cookies received:</h2>', "'onclose' =")

    def testCase17(self):
        url = self.getTestCaseUrl(17, 'TestIMS passed')
        r = self.testApp.get(url)
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>TestIMS</title>',
            '<h2>Test If-Modified-Since support in Webware</h2>',
            'test requires a running web server')
        url = '/PSP/Examples/psplogo.png'
        r = self.testApp.get(url)
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.headers.get('Content-Type'), 'image/png')
        size = len(r.body)
        self.assertGreater(size, 0)
        lastMod = r.headers.get('Last-Modified')
        self.assertTrue(lastMod)
        r = self.testApp.get(url, headers={'If-Modified-Since': lastMod})
        self.assertEqual(r.status, '304 Not Modified')
        self.assertNotIn('Content-Type', r.headers)
        self.assertEqual(len(r.body), 0)
        arpaFormat = '%a, %d %b %Y %H:%M:%S GMT'
        t = time.strptime(lastMod, arpaFormat)
        t = (t[0] - 1,) + t[1:]  # one year before last modification
        beforeMod = time.strftime(arpaFormat, time.gmtime(time.mktime(t)))
        r = self.testApp.get(url, headers={'If-Modified-Since': beforeMod})
        self.assertEqual(r.status, '200 OK')
        self.assertEqual(r.headers.get('Content-Type'), 'image/png')
        self.assertEqual(r.headers.get('Last-Modified'), lastMod)
        self.assertEqual(len(r.body), size)

    def testCase18(self):
        url = self.getTestCaseUrl(18, 'Servlet import test passed')
        r = self.testApp.get(url)
        self.assertEqual(r.status, '200 OK')
        r.mustcontain(
            '<title>ServletImport</title>',
            '<h2>Webware Servlet Import Test</h2>',
            '<h3>Test of servlet import details.</h3>',
            '<p>modName = <code>Testing.ServletImport</code></p>',
            '<p>modNameFromClass = <code>Testing.ServletImport</code></p>',
            'passed', no='failed')


class TestTestingWithExtraPathInfo(TestTesting):

    settings = dict(
        PrintConfigAtStartUp=False,
        ExtraPathInfo=True
    )



import os
import sys


class WillNotRunError(Exception):
    """Error for Webware components that will not run."""


def versionString(version):
    """Convert the given version tuple to a string.

    For a sequence containing version information such as (2, 0, 0, 'b'),
    this returns a printable string such as '2.0b'.
    The micro version number is excluded from the string if it is zero.
    """
    ver = list(map(str, version))
    ver, suffix = ver[:3], '.'.join(ver[3:])
    if len(ver) > 2 and ver[2] == '0':
        ver = ver[:2]
    return '.'.join(ver) + suffix


class PropertiesObject(dict):
    """A Properties Object.

    A PropertiesObject represents, in a dictionary-like fashion, the values
    found in a Properties.py file. That file is always included with a Webware
    component to advertise its name, version, status, etc. Note that a Webware
    component is a Python package that follows additional conventions.
    Also, the top level Webware directory contains a Properties.py.

    Component properties are often used for:
      * generation of documentation
      * runtime examination of components, especially prior to loading

    PropertiesObject provides additional keys:
      * filename - the filename from which the properties were read
      * versionString - a nicely printable string of the version
      * requiredPyVersionString - like versionString,
        but for requiredPyVersion instead
      * willRun - 1 if the component will run.
        So far that means having the right Python version.
      * willNotRunReason - defined only if willRun is 0,
        contains a readable error message

    Using a PropertiesObject is better than investigating the Properties.py
    file directly, because the rules for determining derived keys and any
    future convenience methods will all be provided here.

    Usage example:
        from MiscUtils.PropertiesObject import PropertiesObject
        props = PropertiesObject(filename)
        for key, value in props.items():
            print(f'{key}: {value}')

    Note: We don't normally suffix a class name with "Object" as we have
    with this class, however, the name Properties.py is already used in
    our containing package and all other packages.
    """

    # region Init and reading

    def __init__(self, filename=None):
        dict.__init__(self)
        if filename:
            self.readFileNamed(filename)

    def loadValues(self, *args, **kwargs):
        self.update(*args, **kwargs)
        self.cleanPrivateItems()

    def readFileNamed(self, filename):
        results = {}
        with open(filename) as f:
            exec(f.read(), results)
        self.update(results)
        self.cleanPrivateItems()
        self.createDerivedItems()

    # endregion Init and reading

    # region Self utility

    def cleanPrivateItems(self):
        """Remove items whose keys start with a double underscore."""
        keys = list(self)
        for key in keys:
            if key.startswith('__'):
                del self[key]

    def addBuiltinItems(self):
        pass

    def createDerivedItems(self):
        self.createVersionString()
        self.createRequiredPyVersionString()
        self.createWillRun()

    def createVersionString(self):
        if 'version' in self:
            self['versionString'] = versionString(self['version'])

    def createRequiredPyVersionString(self):
        if 'requiredPyVersion' in self:
            self['requiredPyVersionString'] = versionString(
                self['requiredPyVersion'])

    def createWillRun(self):
        try:
            # Invoke each of the checkFoo() methods
            for key in self.willRunKeys():
                methodName = 'check' + key[0].upper() + key[1:]
                method = getattr(self, methodName)
                method()
        except WillNotRunError as msg:
            self['willRun'] = False
            self['willNotRunReason'] = msg
        else:
            self['willRun'] = True  # we passed all the tests

    def willRunKeys(self):
        """Return keys to be examined before running the component.

        This returns a set of all keys whose values should be examined in
        order to determine if the component will run. Used by createWillRun().
        """
        return {
            'requiredPyVersion', 'requiredOpSys', 'deniedOpSys', 'willRunFunc'}

    def checkRequiredPyVersion(self):
        if 'requiredPyVersion' in self and tuple(
                sys.version_info) < tuple(self['requiredPyVersion']):
            pythonVersion = '.'.join(map(str, sys.version_info))
            requiredVersion = self['requiredPyVersionString']
            raise WillNotRunError(
                f'Required Python version is {requiredVersion},'
                f' but actual version is {pythonVersion}.')

    def checkRequiredOpSys(self):
        requiredOpSys = self.get('requiredOpSys')
        if requiredOpSys:
            # We accept a string or list of strings
            if isinstance(requiredOpSys, str):
                requiredOpSys = [requiredOpSys]
            opSys = os.name
            if opSys not in requiredOpSys:
                requiredOpSys = '/'.join(requiredOpSys)
                raise WillNotRunError(
                    f'Required operating system is {requiredOpSys},'
                    f' but actual operating system is {opSys}.')

    def checkDeniedOpSys(self):
        deniedOpSys = self.get('deniedOpSys')
        if deniedOpSys:
            # We accept a string or list of strings
            if isinstance(deniedOpSys, str):
                deniedOpSys = [deniedOpSys]
            opSys = os.name
            if opSys in deniedOpSys:
                deniedOpSys = '/'.join(deniedOpSys)
                raise WillNotRunError(
                    f'Will not run on operating system {deniedOpSys}'
                    f' and actual operating system is {opSys}.')

    def checkWillRunFunc(self):
        willRunFunc = self.get('willRunFunc')
        if willRunFunc:
            whyNotMsg = willRunFunc()
            if whyNotMsg:
                raise WillNotRunError(whyNotMsg)

    # endregion Self utility



#!/usr/bin/env python3

import sys
import time
from glob import glob
try:
    from cProfile import Profile
except ImportError:
    from profile import Profile

from MiscUtils.DataTable import DataTable


class BenchDataTable:

    def __init__(self, profile=False, runTestSuite=True):
        self._shouldProfile = profile
        self._shouldRunTestSuite = runTestSuite
        self._iterations = 1000

    def main(self):
        print('Benchmarking DataTable ...')
        print()
        if len(sys.argv) > 1 and sys.argv[1].lower().startswith('prof'):
            self._shouldProfile = True
        if self._shouldRunTestSuite:
            print('Running test suite ...')
            from unittest import main
            main('MiscUtils.Tests.TestDataTable', exit=False)
        start = time.time()
        if self._shouldProfile:
            prof = Profile()
            prof.runcall(self._main)
            filename = f'{self.__class__.__name__}.pstats'
            prof.dump_stats(filename)
            print('Wrote', filename)
        else:
            self._main()
        duration = time.time() - start
        print(f'{duration:.1f} secs')

    def _main(self):
        print()
        for name in glob('Sample*.csv'):
            print("Benchmark using", name, "...")
            self.benchFileNamed(name)

    def benchFileNamed(self, name):
        with open(name) as f:
            contents = f.read()
        for _iteration in range(self._iterations):
            # we duplicate lines to reduce the overhead of the loop
            dt = DataTable()
            dt.readString(contents)
            dt = DataTable()
            dt.readString(contents)
            dt = DataTable()
            dt.readString(contents)
            dt = DataTable()
            dt.readString(contents)
            dt = DataTable()
            dt.readString(contents)
            dt = DataTable()
            dt.readString(contents)
            dt = DataTable()
            dt.readString(contents)
            dt = DataTable()
            dt.readString(contents)


if __name__ == '__main__':
    BenchDataTable().main()



from .AdminPage import AdminPage


class AdminSecurity(AdminPage):

    # Set this to False if you want to allow everyone to access secure pages
    # with no login required. This should instead come from a config file.
    _requireLogin = True

    def writeHTML(self):
        session = self.session()
        request = self.request()
        trans = self.transaction()
        app = self.application()
        # Are they logging in?
        if (request.hasField('login')
                and request.hasField('username')
                and request.hasField('password')):
            # They are logging in. Get login id and clear session:
            loginId = session.value('loginId', None)
            session.values().clear()
            # Check if this is a valid user/password
            username = request.field('username')
            password = request.field('password')
            if (self.isValidUserAndPassword(username, password)
                    and request.field('loginId', 'noLogin') == loginId):
                # Success; log them in and send the page:
                session.setValue('authenticated_user_admin', username)
                AdminPage.writeHTML(self)
            else:
                # Failed login attempt; have them try again:
                request.fields()['extra'] = (
                    'Login failed. Please try again.'
                    ' (And make sure cookies are enabled.)')
                app.forward(trans, 'LoginPage')
        # Are they logging out?
        elif request.hasField('logout'):
            # They are logging out. Clear all session variables:
            session.values().clear()
            request.fields()['extra'] = 'You have been logged out.'
            app.forward(trans, 'LoginPage')
        # Are they already logged in?
        elif not self._requireLogin or session.value(
                'authenticated_user_admin', None):
            # They are already logged in; write the HTML for this page:
            AdminPage.writeHTML(self)
        else:
            # They need to log in.
            app.forward(trans, 'LoginPage')

    def isValidUserAndPassword(self, username, password):
        # Replace this with a database lookup, or whatever you're using
        # for authentication...
        adminPassword = self.application().setting('AdminPassword')
        return (username == 'admin' and adminPassword and
                password == adminPassword)



"""Test Webware make script"""

import os
import shutil
import subprocess
import tempfile
import unittest


class TestMakeApp(unittest.TestCase):

    def setUp(self):
        self.testDir = tempfile.mkdtemp()
        self.currentDir = os.getcwd()
        os.chdir(self.testDir)

    def tearDown(self):
        os.chdir(self.currentDir)
        shutil.rmtree(self.testDir)

    def runMake(self, opts=None):
        args = ['webware', 'make']
        if opts:
            args += opts
        result = subprocess.run(
            args, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            check=False, universal_newlines=True)
        self.assertEqual(result.returncode, 0)
        self.assertEqual(result.stderr, '')
        return result.stdout.splitlines()

    def testMakeHelp(self):
        output = self.runMake(['-h'])
        self.assertEqual(output.pop(0),
                         'usage: webware make [-h]'
                         ' [-c CONTEXT_NAME] [-d CONTEXT_DIR] [-l LIBRARY]')
        if hasattr(shutil, 'chown'):
            self.assertEqual(output.pop(0).lstrip(), '[-u USER] [-g GROUP]')
        self.assertEqual(output.pop(0).lstrip(), 'WORK_DIR')

    def testMakeNewApp(self):
        output = self.runMake(['MyApp'])
        self.assertEqual(
            output[0], 'Making a new Webware runtime directory...')
        self.assertIn('Creating the directory tree...', output)
        self.assertIn('Copying config files...', output)
        self.assertIn('Copying other files...', output)
        self.assertIn('Creating default context...', output)
        self.assertIn('Updating config for default context...', output)
        self.assertIn('Creating .gitignore file...', output)
        self.assertIn(
            "Congratulations, you've just created a"
            " runtime working directory for Webware.", output)
        self.assertIn(
            'To start the development server, run this command:', output)
        self.assertIn('webware serve', output)
        self.assertTrue(os.path.isdir('MyApp'))
        os.chdir('MyApp')
        appDirs = 'Cache Configs ErrorMsgs Logs MyContext Sessions'.split()
        for appDir in appDirs:
            self.assertTrue(os.path.isdir(appDir), appDir)
        appFiles = '.gitignore error404.html Scripts/WSGIScript.py'.split()
        for appFile in appFiles:
            self.assertTrue(os.path.isfile(appFile), appFile)
        with open('Scripts/WSGIScript.py') as f:
            wsgiScript = f.read().splitlines()
        self.assertEqual(wsgiScript[0], '#!/usr/bin/env python3')
        self.assertIn('from Application import Application', wsgiScript)
        os.chdir('MyContext')
        with open('__init__.py') as f:
            initScript = f.read().splitlines()
        self.assertIn('def contextInitialize(application, path):', initScript)
        with open('Main.py') as f:
            mainServlet = f.read().splitlines()
        self.assertIn('class Main(Page):', mainServlet)



"""An abstract response"""

from time import time

from MiscUtils import AbstractError


class Response:
    """The abstract response class.

    Response is a base class that offers the following:

      * A time stamp (indicating when the response was finished)
      * An output stream

    Response is an abstract class; developers typically use HTTPResponse.
    """

    # region Init

    def __init__(self, trans, strmOut):
        self._strmOut = strmOut
        self._transaction = trans

    # endregion Init

    # region End time

    def endTime(self):
        return self._endTime

    def recordEndTime(self):
        """Record the end time of the response.

        Stores the current time as the end time of the response. This should
        be invoked at the end of deliver(). It may also be invoked by the
        application for those responses that never deliver due to an error.
        """
        self._endTime = time()

    # endregion End time

    # region Output

    def write(self, output):
        raise AbstractError(self.__class__)

    def isCommitted(self):
        raise AbstractError(self.__class__)

    def deliver(self):
        raise AbstractError(self.__class__)

    def reset(self):
        raise AbstractError(self.__class__)

    def streamOut(self):
        return self._strmOut

    # endregion Output

    # region Cleanup

    def clearTransaction(self):
        del self._transaction

    # endregion Cleanup

    # region Exception reports

    _exceptionReportAttrNames = ['endTime']

    def writeExceptionReport(self, handler):
        handler.writeTitle(self.__class__.__name__)
        handler.writeAttrs(self, self._exceptionReportAttrNames)

    # endregion Exception reports



#!/usr/bin/env python3

"""Serve Webware for Python Application using the waitress WSGI server."""

import argparse


def serve(args):
    try:
        from waitress import serve
    except ImportError as e:
        raise RuntimeError('Waitress server is not installed') from e

    if args.browser:
        scheme = args.url_scheme
        host, port = args.host, args.port
        prefix = args.url_prefix.strip('/')
        if prefix:
            prefix += '/'
        url = f'{scheme}://{host}:{port}/{prefix}'

        import time
        import threading
        import webbrowser

        def openBrowser():
            time.sleep(1)
            webbrowser.open(url)

        t = threading.Thread(target=openBrowser)
        t.setDaemon(True)
        t.start()

    if args.reload:
        try:
            import hupper
        except ImportError as e:
            raise RuntimeError(
                'The hupper process monitor is not installed') from e
        if not hupper.is_active():
            print('Running Webware with reloading option...')
            args.browser = args.reload = False
            hupper.start_reloader(
                'webware.Scripts.WaitressServer.serve',
                reload_interval=int(args.reload_interval),
                worker_args=[args])

    development = not args.prod
    from os import environ
    if development:
        environ['WEBWARE_DEVELOPMENT'] = 'true'
    elif 'WEBWARE_DEVELOPMENT' in environ:
        del environ['WEBWARE_DEVELOPMENT']
    try:
        # get application from WSGI script
        with open('Scripts/WSGIScript.py') as f:
            script = f.read()
        # set development flag in the script
        script = script.replace(
            'development =', f'development = {development} #')
        # do not change working directory in the script
        script = script.replace('workDir =', "workDir = '' #")
        scriptVars = {}
        exec(script, scriptVars)
        application = scriptVars['application']
    except Exception as e:
        raise RuntimeError(
            'Cannot find Webware application.\nIs the current directory'
            ' the application working directory?') from e

    print("Waitress serving Webware application...")
    args = vars(args)
    for arg in 'browser reload reload_interval prod'.split():
        del args[arg]
    if args['trusted_proxy_headers']:
        args['trusted_proxy_headers'] = args[
            'trusted_proxy_headers'].split(',')
    if not args['trusted_proxy']:
        if args['trusted_proxy_count'] == 1:
            del args['trusted_proxy_count']
        if not args['trusted_proxy_headers']:
            del args['trusted_proxy_headers']
    serve(application, **args)


def addArguments(parser):
    """Add command line arguments to the given parser."""
    parser.add_argument(
        '-l', '--host',
        help="Hostname or IP address on which to listen",
        default='127.0.0.1',
    )
    parser.add_argument(
        '-p', '--port',
        help="TCP port on which to listen",
        default='8080',
    )
    parser.add_argument(
        '--url-scheme',
        help="Specifies the scheme portion of the URL",
        default='http',
    )
    parser.add_argument(
        '--url-prefix',
        help="Mount the application using this URL prefix",
        default='',
    )
    parser.add_argument(
        '-r', '--reload',
        action='store_true',
        help="Use auto-restart file monitor",
        default=False,
    )
    parser.add_argument(
        '--reload-interval',
        type=int,
        help="Seconds between checking files",
        default=1,
    )
    parser.add_argument(
        '-b', '--browser',
        action='store_true',
        help="Open a web browser to the running app",
    )
    parser.add_argument(
        '--threads',
        type=int,
        help="Number of threads used to process application logic",
        default=4,
    )
    parser.add_argument(
        '--trusted-proxy',
        help="IP address of a trusted peer passing proxy headers",
        default=None,
    )
    parser.add_argument(
        '--trusted-proxy-count',
        type=int,
        help="How many proxies we trust when chained",
        default=1,
    )
    parser.add_argument(
        '--trusted-proxy-headers',
        help="Comma-separated proxy headers we shall trust",
        default=None,
    )
    parser.add_argument(
        '--prod',
        action='store_true',
        help="Do not set development mode",
        default=False,
    )


def main(args=None):
    """Evaluate the command line arguments and call serve()."""
    parser = argparse.ArgumentParser(
        description="Serve a Webware application using waitress")
    addArguments(parser)
    args = parser.parse_args(args)
    serve(args)


if __name__ == '__main__':
    main()



import os

from MiscUtils.DataTable import DataTable

from .AdminSecurity import AdminSecurity


class DumpCSV(AdminSecurity):

    def filename(self):
        """Overridden by subclasses to specify what filename to show."""
        raise NotImplementedError

    def awake(self, transaction):
        AdminSecurity.awake(self, transaction)
        self._filename = self.filename()

    def shortFilename(self):
        return os.path.splitext(os.path.split(self._filename)[1])[0]

    def title(self):
        return 'View ' + self.shortFilename()

    def writeContent(self):
        if not os.path.exists(self._filename):
            self.writeln('<p>File does not exist.</p>')
            return
        table = DataTable(self._filename)
        plural = '' if len(table) == 1 else 's'
        self.writeln(f'<p>{len(table)} row{plural}</p>')
        self.writeln('<table class="NiceTable">')
        # Head row gets special formatting
        self._headings = [col.name().strip() for col in table.headings()]
        self._numCols = len(self._headings)
        self.writeln('<tr>')
        for value in self._headings:
            self.writeln('<th>', value, '</th>')
        self.writeln('</tr>')
        # Data rows
        for rowIndex, row in enumerate(table, 1):
            self.writeln('<tr>')
            for colIndex, value in enumerate(row):
                if colIndex >= self._numCols:
                    break  # skip surplus columns
                self.writeln('<td>',
                             self.cellContents(rowIndex, colIndex, value),
                             '</td>')
            self.writeln('</tr>')
        self.writeln('</table>')

    def cellContents(self, _rowIndex, _colIndex, value):
        """Hook for subclasses to customize the contents of a cell.

        Based on any criteria (including location).
        """
        return value



"""Automated tests for the PSP Context"""

import os
import unittest

from PSP.Context import PSPCLContext


class TestCLContext(unittest.TestCase):

    def testInit(self):
        pspFile = '/files/PSP/ContextTest.psp'.replace('/', os.sep)
        clc = PSPCLContext(pspFile)
        self.assertEqual(clc.getFullPspFileName(), pspFile)
        self.assertEqual(clc.getPspFileName(), 'ContextTest.psp')
        self.assertEqual(clc.getBaseUri(), '/files/PSP'.replace('/', os.sep))

    def testPythonFileEncoding(self):
        clc = PSPCLContext('test.psp')
        self.assertEqual(clc.getPythonFileEncoding(), None)
        clc.setPythonFileEncoding('latin-1')
        self.assertEqual(clc.getPythonFileEncoding(), 'latin-1')

    def testResolveRelativeURI(self):
        pspFile = '/files/PSP/Test1.psp'.replace('/', os.sep)
        clc = PSPCLContext(pspFile)
        uri = clc.resolveRelativeURI('Test2.psp')
        self.assertEqual(uri, '/files/PSP/Test2.psp'.replace('/', os.sep))
        self.assertEqual(clc.resolveRelativeURI(uri), uri)

    def testPSPReader(self):
        reader = object()
        clc = PSPCLContext('test.psp')
        clc.setPSPReader(reader)
        self.assertEqual(clc.getReader(), reader)

    def testClassName(self):
        clc = PSPCLContext('test.psp')
        clc.setClassName('ContextTestClass')
        self.assertEqual(clc.getServletClassName(), 'ContextTestClass')



"""HTTP responses"""

from datetime import datetime, timedelta
from time import time, gmtime, strftime, struct_time

from MiscUtils import NoDefault
from MiscUtils.DateInterval import timeDecode
from MiscUtils.Funcs import localTimeDelta
from WebUtils.Funcs import htmlEncode

from Response import Response
from Cookie import Cookie  # pylint: disable=wrong-import-order
from HTTPExceptions import HTTPException, HTTPServerError

debug = False


class HTTPResponse(Response):
    """The base class for HTTP responses."""

    # region Init

    def __init__(self, transaction, strmOut, headers=None):
        """Initialize the request."""
        Response.__init__(self, transaction, strmOut)
        self._committed = False
        if headers is None:
            self._headers = {}
            self.setHeader('Content-Type', 'text/html')
        else:
            self._headers = headers
        self._cookies = {}

    # endregion Init

    # region Protocol

    def protocol(self):
        """Return the name and version of the protocol."""
        return self._transaction.request().protocol()

    # endregion Protocol

    # region Headers

    def header(self, name, default=NoDefault):
        """Return the value of the specified header."""
        if default is NoDefault:
            return self._headers[name.capitalize()]
        return self._headers.get(name.capitalize(), default)

    def hasHeader(self, name):
        return name.capitalize() in self._headers

    def setHeader(self, name, value):
        """Set a specific header by name.

        Parameters:
            name: the header name
            value: the header value
        """
        self.assertNotCommitted()
        self._headers[name.capitalize()] = value

    def delHeader(self, name):
        """Delete a specific header by name."""
        self.assertNotCommitted()
        del self._headers[name.capitalize()]

    def headers(self):
        """Return all the headers.

        Returns a dictionary-style object of all header objects contained by
        this request.
        """
        return self._headers

    def clearHeaders(self):
        """Clear all the headers.

        You might consider a setHeader('Content-Type', 'text/html')
        or something similar after this.
        """
        self.assertNotCommitted()
        self._headers = {}

    # endregion Headers

    # region Cookies

    def cookie(self, name):
        """Return the value of the specified cookie."""
        return self._cookies[name]

    def hasCookie(self, name):
        """Return True if the specified cookie is present."""
        return name in self._cookies

    def setCookie(
            self, name, value, path='/', expires='ONCLOSE', secure=False):
        """Set a cookie.

        You can also set the path (which defaults to /).

        You can also set when it expires. It can expire:

        - 'NOW': this is the same as trying to delete it, but it
          doesn't really seem to work in IE
        - 'ONCLOSE': the default behavior for cookies (expires when
          the browser closes)
        - 'NEVER': some time in the far, far future.
        - integer: a timestamp value
        - tuple or struct_time: a tuple, as created by the time module
        - datetime: a datetime.datetime object for the time (if without
          time zone, assumed to be *local*, not GMT time)
        - timedelta: a duration counted from the present, e.g.,
          datetime.timedelta(days=14) (2 weeks in the future)
        - '+...': a time in the future, '...' should be something like
          1w (1 week), 3h46m (3:45), etc.  You can use y (year),
          b (month), w (week), d (day), h (hour), m (minute),
          s (second). This is done by the MiscUtils.DateInterval.
        """
        cookie = Cookie(name, value)
        t = expires
        if isinstance(t, str):
            if t == 'ONCLOSE':
                t = None
            elif t == 'NOW':
                cookie.delete()
                return
            elif t == 'NEVER':
                t = gmtime()
                t = (t[0] + 10,) + t[1:]
            elif t.startswith('+'):
                t = time() + timeDecode(t[1:])
        if t:
            if isinstance(t, (int, float)):
                t = gmtime(t)
            if isinstance(t, (tuple, struct_time)):
                t = strftime("%a, %d-%b-%Y %H:%M:%S GMT", t)
            if isinstance(t, timedelta):
                t = datetime.now() + t
            if isinstance(t, datetime):
                d = t.utcoffset()
                if d is None:
                    d = localTimeDelta()
                t -= d
                t = t.strftime("%a, %d-%b-%Y %H:%M:%S GMT")
            cookie.setExpires(t)
        if path:
            cookie.setPath(path)
        if secure:
            cookie.setSecure(secure)
        self.addCookie(cookie)

    def addCookie(self, cookie):
        """Add a cookie that will be sent with this response.

        cookie is a Cookie object instance. See the Cookie class docs.
        """
        self.assertNotCommitted()
        if not isinstance(cookie, Cookie):
            raise TypeError(f'Expected a Cookie object, but got: {cookie!r}')
        self._cookies[cookie.name()] = cookie

    def delCookie(self, name, path='/', secure=False):
        """Delete a cookie at the browser.

        To do so, one has to create and send to the browser a cookie with
        parameters that will cause the browser to delete it.
        """
        if name in self._cookies:
            self._cookies[name].delete()
        else:
            cookie = Cookie(name, None)
            if path:
                cookie.setPath(path)
            if secure:
                cookie.setSecure(secure)
            cookie.delete()
            self.addCookie(cookie)

    def cookies(self):
        """Get all the cookies.

        Returns a dictionary-style object of all Cookie objects that will
        be sent with this response.
        """
        return self._cookies

    def clearCookies(self):
        """Clear all the cookies."""
        self.assertNotCommitted()
        self._cookies = {}

    # endregion Cookies

    # region Status

    def setStatus(self, code, msg=''):
        """Set the status code of the response, such as 200, 'OK'."""
        self.assertNotCommitted()
        self.setHeader('Status', f'{code} {msg}')

    # endregion Status

    # region Special responses

    def sendError(self, code, msg=''):
        """Set the status code to the specified code and message."""
        self.assertNotCommitted()
        self.setStatus(code, msg)

    def sendRedirect(self, url, status=None):
        """Redirect to another url.

        This method sets the headers and content for the redirect,
        but does not change the cookies and other headers.
        Use clearCookies() or clearHeaders() as appropriate.

        See https://www.ietf.org/rfc/rfc2616 (section 10.3.3)
        and https://www.ietf.org/rfc/rfc3875 (section 6.2.3).
        """
        self.assertNotCommitted()
        self.setHeader('Status', status or '302 Found')
        self.setHeader('Location', url)
        self.setHeader('Content-Type', 'text/html')
        self.write(
            '<html><body>This page has been redirected'
            ' to <a href="{0}">{0}</a>.</body></html>'.format(htmlEncode(url)))

    def sendRedirectPermanent(self, url):
        """Redirect permanently to another URL."""
        self.sendRedirect(url, status='301 Moved Permanently')

    def sendRedirectSeeOther(self, url):
        """Redirect to a URL that shall be retrieved with GET.

        This method exists primarily to allow for the PRG pattern.

        See https://en.wikipedia.org/wiki/Post/Redirect/Get
        """
        self.sendRedirect(url, status='303 See Other')

    def sendRedirectTemporary(self, url):
        """Redirect temporarily to another URL."""
        self.sendRedirect(url, status='307 Temporary Redirect')

    # endregion Special responses

    # region Output

    def write(self, output=None):
        """Write output to the response stream.

        The output will be converted to a string, and then converted to bytes
        using the application output encoding, unless it is already bytes.
        """
        if output:
            self._strmOut.write(output)
        if not self._committed and self._strmOut._needCommit:
            self.commit()

    def flush(self, autoFlush=True):
        """Send all accumulated response data now.

        Commits the response headers and tells the underlying stream to flush.
        if autoFlush is true, the responseStream will flush itself
        automatically from now on.

        Caveat: Some web servers, especially IIS, will still buffer the output
        from your servlet until it terminates before transmitting the results
        to the browser. Also, server modules for Apache like mod_deflate or
        mod_gzip may do buffering of their own that will cause flush() to not
        result in data being sent immediately to the client. You can prevent
        this by setting a no-gzip note in the Apache configuration, e.g.

           SetEnvIf Request_URI ^/Webware/MyServlet no-gzip=1

        Even the browser may buffer its input before displaying it.
        For example, Netscape buffered text until it received an end-of-line
        or the beginning of a tag, and it didn't render tables until the end
        tag of the outermost table was seen. Some Firefox add-ons also buffer
        response data before it gets rendered. Some versions of MSIE will only
        start to display the page after they have received 256 bytes of output,
        so you may need to send extra whitespace before flushing to get MSIE
        to display the page.
        """
        if not self._committed:
            self.commit()
        self._strmOut.flush()
        self._strmOut.setAutoCommit(autoFlush)

    def isCommitted(self):
        """Check whether response is already committed.

        Checks whether the response has already been partially or completely
        sent. If this method returns True, then no new headers/cookies can be
        added to the response anymore.
        """
        return self._committed

    def assertNotCommitted(self):
        """Assert the the response is not already committed.

        This raises a ConnectionError if the connection is already committed.
        """
        if self._committed:
            raise ConnectionError("Headers have already been sent.")

    def deliver(self):
        """Deliver response.

        The final step in the processing cycle.
        Not used for much with responseStreams added.
        """
        if debug:
            print("HTTPResponse deliver called")
        self.recordEndTime()
        if not self._committed:
            self.commit()

    def commit(self):
        """Commit response.

        Write out all headers to the response stream, and tell the underlying
        response stream it can start sending data.
        """
        if debug:
            print("HTTPResponse commit")
        self.recordSession()
        if self._transaction.errorOccurred():
            err = self._transaction.error()
            if not isinstance(err, HTTPException):
                err = HTTPServerError()
                self._transaction.setError(err)
            self.setErrorHeaders(err)
        self.writeHeaders()
        self._committed = True
        self._strmOut.commit()

    def writeHeaders(self):
        """Write headers to the response stream. Used internally."""
        if self._committed:
            print("response.writeHeaders called when already committed")
            return
        # make sure the status header comes first
        if 'Status' in self._headers:
            # store and temporarily delete status
            status = self._headers['Status']
            del self._headers['Status']
        else:
            # invent meaningful status
            status = '302 Found' if 'Location' in self._headers else '200 OK'
        headers = list(self._headers.items())
        self._headers['Status'] = status  # restore status
        headers.extend(
            ('Set-Cookie', c.headerValue()) for c in self._cookies.values())
        self._strmOut.startResponse(status, headers)

    def recordSession(self):
        """Record session ID.

        Invoked by commit() to record the session ID in the response
        (if a session exists). This implementation sets a cookie for
        that purpose. For people who don't like sweets, a future version
        could check a setting and instead of using cookies, could parse
        the HTML and update all the relevant URLs to include the session ID
        (which implies a big performance hit). Or we could require site
        developers to always pass their URLs through a function which adds
        the session ID (which implies pain). Personally, I'd rather just
        use cookies. You can experiment with different techniques by
        subclassing Session and overriding this method. Just make sure
        Application knows which "session" class to use.

        It should be also considered to automatically add the server port
        to the cookie name in order to distinguish application instances
        running on different ports on the same server, or to use the port
        cookie-attribute introduced with RFC 2965 for that purpose.
        """
        trans = self._transaction
        app = trans.application()
        if not app.setting('UseCookieSessions'):
            return
        session = trans._session
        if not session:
            if debug:
                print('>> recordSession: Did not set SID.')
            return
        request = trans.request()
        sessionName = app.sessionName(trans)
        identifier = session.identifier()
        if session.isExpired() or session.timeout() == 0:
            self.delCookie(
                sessionName, app.sessionCookiePath(trans),
                request.isSecure() and app.setting('SecureSessionCookie'))
            if debug:
                print('>> recordSession: Removing SID', identifier)
            return
        if request.hasCookie(sessionName):
            if request.cookie(sessionName) == identifier:
                if debug:
                    print('>> recordSession: Using SID', identifier)
                return
        cookie = Cookie(app.sessionName(trans), identifier)
        cookie.setPath(app.sessionCookiePath(trans))
        if trans.request().isSecure():
            cookie.setSecure(app.setting('SecureSessionCookie'))
        self.addCookie(cookie)
        if debug:
            print('>> recordSession: Setting SID', identifier)

    def reset(self):
        """Reset the response (such as headers, cookies and contents)."""
        self.assertNotCommitted()
        self._headers = {}
        self.setHeader('Content-Type', 'text/html')
        self._cookies = {}
        self._strmOut.clear()

    def rawResponse(self):
        """Return the final contents of the response.

        Don't invoke this method until after deliver().

        Returns a dictionary representing the response containing only
        strings, numbers, lists, tuples, etc. with no backreferences.
        That means you don't need any special imports to examine the contents
        and you can marshal it. Currently there are two keys. 'headers' is
        list of tuples each of which contains two strings: the header and
        it's value. 'contents' is a string (that may be binary, for example,
        if an image were being returned).
        """
        headers = []
        for key, value in self._headers.items():
            headers.append((key, value))
        for cookie in self._cookies.values():
            headers.append(('Set-Cookie', cookie.headerValue()))
        return dict(headers=headers, contents=self._strmOut.buffer())

    def size(self):
        """Return the size of the final contents of the response.

        Don't invoke this method until after deliver().
        """
        return self._strmOut.size()

    def mergeTextHeaders(self, headerStr):
        """Merge text into our headers.

        Given a string of headers (separated by newlines),
        merge them into our headers.
        """
        lines = headerStr.splitlines()
        for line in lines:
            header = line.split(':', 1)
            if len(header) > 1:
                self.setHeader(header[0], header[1].rstrip())

    # endregion Output

    # region Exception reports

    _exceptionReportAttrNames = Response._exceptionReportAttrNames + [
        'committed', 'headers', 'cookies']

    def setErrorHeaders(self, err):
        """Set error headers for an HTTPException."""
        for header, value in err.headers().items():
            self.setHeader(header, value)
        self.setStatus(err.code(), err.codeMessage())
        self.setHeader('Content-Type', 'text/html')

    def displayError(self, err):
        """Display HTTPException errors, with status codes."""
        self.assertNotCommitted()
        self._strmOut.clear()
        self._strmOut.write(err.html())
        uri = self._transaction.request().uri()
        print(f'HTTPResponse: {uri}: {err.codeMessage()}')
        self.commit()

    # endregion Exception reports



"""This module handles requests from the application for PSP pages.

Copyright (c) by Jay Love, 2000 (mailto:jsliv@jslove.org)

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee or royalty is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation or portions thereof, including modifications,
that you make.
"""

import sys
from os import listdir, unlink, utime
from os.path import (
    getatime, getmtime, exists, isdir, isfile, join, normcase, splitdrive)
from shutil import rmtree
from string import digits, ascii_letters

from ServletFactory import ServletFactory
from . import Context, PSPCompiler


class PSPServletFactory(ServletFactory):
    """Servlet Factory for PSP files."""

    def __init__(self, application):
        ServletFactory.__init__(self, application)
        self._cacheDir = join(application._cacheDir, 'PSP')
        sys.path.append(self._cacheDir)
        self._cacheClassFiles = self._cacheClasses
        t = ['_'] * 256
        for c in digits + ascii_letters:
            t[ord(c)] = c
        self._classNameTrans = ''.join(t)
        setting = application.setting
        self._extensions = setting('ExtensionsForPSP', ['.psp'])
        self._fileEncoding = setting('PSPFileEncoding', None)
        if setting('ClearPSPCacheOnStart', False):
            self.clearFileCache()
        self._recordFile = application._imp.recordFile

    def uniqueness(self):
        return 'file'

    def extensions(self):
        return self._extensions

    def fileEncoding(self):
        """Return the file encoding used in PSP files."""
        return self._fileEncoding

    def flushCache(self):
        """Clean out the cache of classes in memory and on disk."""
        ServletFactory.flushCache(self)
        self.clearFileCache()

    def clearFileCache(self):
        """Clear class files stored on disk."""
        cacheDir = self._cacheDir
        for filename in listdir(cacheDir):
            path = join(cacheDir, filename)
            if isfile(path):  # remove the source files
                unlink(path)
            elif isdir(path):  # also remove the __pycache__
                rmtree(path)

    def computeClassName(self, pageName):
        """Generates a (hopefully) unique class/file name for each PSP file.

        Argument: pageName: the path to the PSP source file
        Returns: a unique name for the class generated fom this PSP source file
        """
        # Compute class name by taking the normalized path and substituting
        # underscores for all non-alphanumeric characters:
        normName = normcase(splitdrive(pageName)[1])
        return normName.translate(self._classNameTrans)

    def loadClassFromFile(self, transaction, fileName, className):
        """Create an actual class instance.

        The module containing the class is imported as though it were a
        module within the context's package (and appropriate subpackages).
        """
        module = self.importAsPackage(transaction, fileName)
        try:
            return getattr(module, className)
        except AttributeError:
            raise AttributeError(
                'Cannot find expected class'
                f' named {className} in {fileName}.') from None

    def loadClass(self, transaction, path):
        className = self.computeClassName(path)
        classFile = join(self._cacheDir, className + ".py")
        mtime = getmtime(path)
        if not exists(classFile) or getmtime(classFile) != mtime:
            context = Context.PSPCLContext(path)
            context.setClassName(className)
            context.setPythonFileName(classFile)
            context.setPythonFileEncoding(self._fileEncoding)
            clc = PSPCompiler.Compiler(context)
            sourceFiles = clc.compile()
            # Set the modification time of the compiled file
            # to be the same as the source file;
            # that's how we'll know if it needs to be recompiled:
            utime(classFile, (getatime(classFile), mtime))
            # Record all included files so we can spot any changes:
            for sourcefile in sourceFiles:
                self._recordFile(sourcefile)
        return self.loadClassFromFile(transaction, classFile, className)



"""Automated tests for PSPUtils

Copyright (c) by Winston Wolff, 2004 https://www.stratolab.com
"""

import unittest

from PSP import PSPUtils


class TestUtils(unittest.TestCase):

    def testRemoveQuotes(self):
        self.assertEqual(PSPUtils.removeQuotes(
            r'html <% psp %\\> html'), 'html <% psp %> html')

    def testIsExpression(self):
        self.assertEqual(PSPUtils.isExpression('<%= bla'), False)
        self.assertEqual(PSPUtils.isExpression('bla %>'), False)
        self.assertEqual(PSPUtils.isExpression('<%= bla %>'), True)
        self.assertEqual(PSPUtils.isExpression('bla <%= bla %>'), False)
        self.assertEqual(PSPUtils.isExpression('<%= bla %> bla'), False)

    def testGetExpr(self):
        self.assertEqual(PSPUtils.getExpr('<%= bla'), '')
        self.assertEqual(PSPUtils.getExpr('bla %>'), '')
        self.assertEqual(PSPUtils.getExpr('<%= bla %>'), ' bla ')
        self.assertEqual(PSPUtils.getExpr('bla <%= bla %>'), '')
        self.assertEqual(PSPUtils.getExpr('<%= bla %> bla'), '')

    def testSplitLines(self):
        self.assertEqual(PSPUtils.splitLines(
            'foo\nbar\n'), ['foo', 'bar'])
        self.assertEqual(PSPUtils.splitLines(
            'foo\rbar\r'), ['foo', 'bar'])
        self.assertEqual(PSPUtils.splitLines(
            'foo\rbar\n', True), ['foo\r', 'bar\n'])

    def testStartsNewBlock(self):
        startsNewBlock = PSPUtils.startsNewBlock
        self.assertEqual(startsNewBlock('x = 1'), False)
        self.assertEqual(startsNewBlock('if x == 1:'), True)
        self.assertEqual(startsNewBlock('x = 1 # bla:'), False)
        self.assertEqual(startsNewBlock('if x == 1: # bla'), True)
        self.assertEqual(startsNewBlock('if x == 1: bla'), False)
        self.assertEqual(startsNewBlock('x = "if x == 1:" # bla:'), False)

    def testCheckAttributes(self):
        checkAttributes = PSPUtils.checkAttributes
        for attrs in (dict(man=1), dict(man=1, opt=1)):
            checkAttributes('test', attrs, (['man'], ['opt']))
        PSPParserException = PSPUtils.PSPParserException
        for attrs in (dict(), dict(opt=1), dict(man=1, noopt=1)):
            self.assertRaises(
                PSPParserException, checkAttributes,
                'test', attrs, (['man'], ['opt']))
        self.assertRaises(
            PSPParserException, checkAttributes,
            'test', dict(opt=1), (['man1', 'man2'], []))
        self.assertRaises(
            PSPParserException, checkAttributes,
            'test', dict(man=1), ([], ['opt1', 'opt2']))

    def testNormalizeIndentation(self):
        normalizeIndentation = PSPUtils.normalizeIndentation

        before = """
    def add(a,b):
      return a+b"""
        expected = """
def add(a,b):
  return a+b"""
        self.assertEqual(normalizeIndentation(before), expected)

        # Comments should be ignored for the un-indentation
        before = """
# Will comments throw off the indentation?
    def add(a,b):
      return a+b"""
        expected = """
# Will comments throw off the indentation?
def add(a,b):
  return a+b"""
        self.assertEqual(normalizeIndentation(before), expected)

        # Will blank lines cause a problem?
        before = """
# Will blank lines cause a problem?

    def add(a,b):

      return a+b"""
        expected = """
# Will blank lines cause a problem?

def add(a,b):

  return a+b"""
        self.assertEqual(normalizeIndentation(before), expected)

        # Tab chars OK?
        before = '\tdef add(a,b):\n\t\treturn a+b'
        expected = 'def add(a,b):\n\treturn a+b'
        self.assertEqual(normalizeIndentation(before), expected)

        # Different line endings OK?
        before = '#line endings\r  def add(a,b):\r  \r  return a+b'
        expected = '#line endings\rdef add(a,b):\r\rreturn a+b'
        self.assertEqual(normalizeIndentation(before), expected)



#
# Ajax "Suggest" Example
#
# Written by Robert Forkel based on
# https://webwareforpython.github.io/w4py-olde-wiki/ajaxinwebware.html
# and http://www.dynamicajax.com/fr/AJAX_Suggest_Tutorial-271_290_312.html,
# with minor changes made by Christoph Zwerschke.
#

from random import randint

from .AjaxPage import AjaxPage

maxSuggestions = 10

# Create some random "magic words":
maxLetters, maxWords = 5, 5000
magicWords = [''.join(chr(randint(97, 122)) for j in range(maxLetters))
              for i in range(maxWords)]


class AjaxSuggest(AjaxPage):

    _clientPolling = None  # we have no long-running queries

    def writeJavaScript(self):
        AjaxPage.writeJavaScript(self)
        self.writeln(
            '<script src="ajaxsuggest.js"></script>')

    def writeStyleSheet(self):
        AjaxPage.writeStyleSheet(self)
        self.writeln(
            '<link rel="stylesheet" href="ajaxsuggest.css">')

    def htBodyArgs(self):
        return AjaxPage.htBodyArgs(self) + ' onload="initPage();"'

    def writeContent(self):
        self.writeln('<h2>Ajax "Suggest" Example</h2>')
        if self.request().hasField('query'):
            query = self.htmlEncode(self.request().field('query'))
            self.writeln(f'''
<p>You have just entered the word <b class="in_red">"{query}"</b>.</p>
<p>If you like, you can try again:</p>''')
        else:
            self.writeln('''
<p>This example uses Ajax techniques to make suggestions
based on your input as you type.</p>
<p>Of course, you need a modern web browser with
JavaScript enabled in order for this to work.</p>
<p>Start typing in some lowercase letters,
and get random words starting with these characters suggested:</p>''')
        self.writeln('''<form><div>
<input type="text" name="query" id="query"
 onkeyup="getSuggestions();" autocomplete="off">
<input type="submit" value="Submit"></div><div class="hide" id="suggestions">
</div></form>''')

    def exposedMethods(self):
        """Register the suggest method for use with Ajax."""
        return ['suggest']

    def suggest(self, prefix):
        """We return a JavaScript function call as string.

        The JavaScript function we want called is `handleSuggestions`
        and we pass an array of strings starting with prefix.

        Note: to pass more general Python objects to the client side,
        use Python's JSON encoder.
        """
        words = [w for w in magicWords if w.startswith(prefix)] or ['none']
        wordList = ",".join(map(repr, words[:maxSuggestions]))
        return f"handleSuggestions([{wordList}]);"



"""Mock python-memcached module."""

import sys
from copy import copy

sys.modules['memcache'] = sys.modules[__name__]

data = {}  # our mock memcache


# pylint: disable=unused-argument, invalid-name

class Client:
    """Mock memcache client."""

    def __init__(self, servers, debug=0, pickleProtocol=0):
        self._connected = True

    def set(self, key, val, time=0):
        if self._connected:
            if val is not None:
                data[key] = val
            return 1
        return 0

    def get(self, key):
        if self._connected:
            return copy(data.get(key))

    def delete(self, key, time=0):
        if self._connected:
            if key in data:
                del data[key]
            return 1
        return 0

    def disconnect_all(self):
        self._connected = False



"""The standard web page template."""

from WebUtils import Funcs
from HTTPContent import HTTPContent


class Page(HTTPContent):
    """The standard web page template.

    Page is a type of HTTPContent that is more convenient for servlets
    which represent HTML pages generated in response to GET and POST requests.
    In fact, this is the most common type of Servlet.

    Subclasses typically override ``writeHeader``, ``writeBody`` and
    ``writeFooter``.

    They might also choose to override ``writeHTML`` entirely.

    When developing a full-blown website, it's common to create a subclass of
    ``Page`` called ``SitePage`` which defines the common look and feel of the
    website and provides site-specific convenience methods. Then all other
    pages in your application then inherit from ``SitePage``.
    """

    # region Transactions

    def defaultAction(self):
        """The default action in a Page is to writeHTML()."""
        self.writeHTML()

    # endregion Transactions

    # region Generating results

    def title(self):
        """The page title.

        Subclasses often override this method to provide a custom title.
        This title should be absent of HTML tags. This implementation
        returns the name of the class, which is sometimes appropriate
        and at least informative.
        """
        return self.__class__.__name__

    def htTitle(self):
        """The page title as HTML.

        Return self.title(). Subclasses sometimes override this to provide
        an HTML enhanced version of the title. This is the method that should
        be used when including the page title in the actual page contents.
        """
        return self.title()

    def htRootArgs(self):
        """The attributes for the <html> element.

        Returns the arguments used for the root HTML tag.
        Invoked by writeHTML() and preAction().

        Authors are encouraged to specify a lang attribute,
        giving the document's language.
        """
        return 'lang="en"'

    def htBodyArgs(self):
        """The attributes for the <body> element.

        Returns the arguments used for the HTML ``<body>`` tag.
        Invoked by writeBody().

        With the prevalence of stylesheets (CSS), you can probably skip
        this particular HTML feature, but for historical reasons this sets
        the page to black text on white.
        """
        return 'style="color:black;background-color:white"'

    def writeHTML(self):
        """Write all the HTML for the page.

        Subclasses may override this method (which is invoked by ``_respond``)
        or more commonly its constituent methods, ``writeDocType``,
        ``writeHead`` and ``writeBody``.

        You will want to override this method if:
          * you want to format the entire HTML page yourself
          * if you want to send an HTML page that has already been generated
          * if you want to use a template that generates the entire page
          * if you want to send non-HTML content; in this case, be sure to call
            self.response().setHeader('Content-Type', 'mime/type').
        """
        self.writeDocType()
        htmlArgs = self.htRootArgs()
        if htmlArgs:
            self.writeln(f'<html {htmlArgs}>')
        else:
            self.writeln('<html>')
        self.writeHead()
        self.writeBody()
        self.writeln('</html>')

    def writeDocType(self):
        """Write the DOCTYPE tag.

        Invoked by ``writeHTML`` to write the ``<!DOCTYPE ...>`` tag.

        By default this gives the HTML 5 DOCTYPE.

        Subclasses may override to specify something else.
        """
        self.writeln('<!DOCTYPE html>')

    def writeHead(self):
        """Write the <head> element of the page.

        Writes the ``<head>`` portion of the page by writing the
        ``<head>...</head>`` tags and invoking ``writeHeadParts`` in between.
        """
        wr = self.writeln
        wr('<head>')
        self.writeHeadParts()
        wr('</head>')

    def writeHeadParts(self):
        """Write the parts included in the <head> element.

        Writes the parts inside the ``<head>...</head>`` tags.
        Invokes ``writeTitle`` and then ``writeMetaData``, ``writeStyleSheet``
        and ``writeJavaScript``. Subclasses should override the ``title``
        method and the three latter methods only.
        """
        self.writeTitle()
        self.writeMetaData()
        self.writeStyleSheet()
        self.writeJavaScript()

    def writeTitle(self):
        """Write the <title> element of the page.

        Writes the ``<title>`` portion of the page.
        Uses ``title``, which is where you should override.
        """
        self.writeln(f'\t<title>{self.title()}</title>')

    def writeMetaData(self):
        """Write the meta data for the page.

        This default implementation only specifies the output encoding.
        Subclasses should override if necessary.
        """
        charset = self.outputEncoding()
        if charset:
            self.writeln(f'\t<meta charset="{charset}">')

    def writeStyleSheet(self):
        """Write the CSS for the page.

        This default implementation does nothing.
        Subclasses should override if necessary.

        A typical implementation is::

            self.writeln('<link rel="stylesheet" href="StyleSheet.css">')
        """
        # base method does nothing

    def writeJavaScript(self):
        """Write the JavaScript for the page.

        This default implementation does nothing.
        Subclasses should override if necessary.

        A typical implementation is::

            self.writeln('<script src="ajax.js"></script>')
        """
        # base method does nothing

    def writeBody(self):
        """Write the <body> element of the page.

        Writes the ``<body>`` portion of the page by writing the
        ``<body>...</body>`` (making use of ``htBodyArgs``) and
        invoking ``writeBodyParts`` in between.
        """
        wr = self.writeln
        bodyArgs = self.htBodyArgs()
        if bodyArgs:
            wr(f'<body {bodyArgs}>')
        else:
            wr('<body>')
        self.writeBodyParts()
        wr('</body>')

    def writeBodyParts(self):
        """Write the parts included in the <body> element.

        Invokes ``writeContent``. Subclasses should only override this method
        to provide additional page parts such as a header, sidebar and footer,
        that a subclass doesn't normally have to worry about writing.

        For writing page-specific content, subclasses should override
        ``writeContent`` instead. This method is intended to be overridden
        by your SitePage.

        See ``SidebarPage`` for an example override of this method.

        Invoked by ``writeBody``.
        """
        self.writeContent()

    def writeContent(self):
        """Write the unique, central content for the page.

        Subclasses should override this method (not invoking super) to write
        their unique page content.

        Invoked by ``writeBodyParts``.
        """
        self.writeln('<p>This page has not yet customized its content.</p>')

    def preAction(self, actionName):
        """Things to do before actions.

        For a page, we first writeDocType(), <html>, and then writeHead().
        """
        self.writeDocType()
        htmlArgs = self.htRootArgs()
        if htmlArgs:
            self.writeln(f'<html {htmlArgs}>')
        else:
            self.writeln('<html>')
        self.writeHead()

    def postAction(self, actionName):
        """Things to do after actions.

        Simply close the html tag (</html>).
        """
        self.writeln('</html>')

    # endregion Generating results

    # region Convenience Methods

    @staticmethod
    def htmlEncode(s):
        """HTML encode special characters.
        Alias for ``WebUtils.Funcs.htmlEncode``, quotes the special characters
        &, <, >, and \"
        """
        return Funcs.htmlEncode(s)

    @staticmethod
    def htmlDecode(s):
        """HTML decode special characters.

        Alias for ``WebUtils.Funcs.htmlDecode``. Decodes HTML entities.
        """
        return Funcs.htmlDecode(s)

    # endregion Convenience Methods



import unittest


class HTMLStatusCodesTest(unittest.TestCase):

    def testHTTPStatusCodeList(self):
        from WebUtils.HTTPStatusCodes import HTTPStatusCodeList
        self.assertIsInstance(HTTPStatusCodeList, list)
        self.assertIn(
            (409, 'CONFLICT', 'Request conflict'), HTTPStatusCodeList)

    def testHTTPStatusCodes(self):
        from WebUtils.HTTPStatusCodes import HTTPStatusCodes
        self.assertIsInstance(HTTPStatusCodes, dict)
        d = HTTPStatusCodes[409]
        self.assertIs(d, HTTPStatusCodes['CONFLICT'])
        description = d['description']
        self.assertEqual(description, 'Request conflict')
        self.assertEqual(d['asciiMsg'], description)
        self.assertEqual(d['htmlMsg'], description)

    def testHTMLTableOfHTTPStatusCodes(self):
        from WebUtils.HTTPStatusCodes import htmlTableOfHTTPStatusCodes
        table = htmlTableOfHTTPStatusCodes()
        self.assertTrue(table.startswith('<table>'))
        self.assertIn('<th>Identifier</th>', table)
        self.assertIn(
            '<tr style="vertical-align:top"><td>404</td><td>NOT_FOUND</td>'
            '<td>Nothing matches the given URI</td></tr>', table)
        self.assertTrue(table.endswith('</table>'))



"""This module co-ordinates the reading of the source file.

It maintains the current position of the parser in the source file.

Copyright (c) by Jay Love, 2000 (mailto:jsliv@jslove.org)

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee or royalty is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation or portions thereof, including modifications,
that you make.

This software is based in part on work done by the Jakarta group.
"""

from copy import copy
import os

from .PSPUtils import PSPParserException


class Mark:
    """This class marks a point in an input stream."""

    def __init__(
            self, reader,
            fileId=None, stream=None, inBaseDir=None, encoding=None):
        if not isinstance(reader, StreamReader):
            raise TypeError('First argument must be a StreamReader')
        self.reader = reader
        self.fileId = fileId
        self.includeStack = []
        self.cursor = 0
        self.stream = stream
        self.baseDir = inBaseDir
        self.encoding = encoding

    def __str__(self):
        return f'{self.getFile()}({self.cursor})'

    def getFile(self):
        return self.reader.getFile(self.fileId)

    def pushStream(self, inFileId, inStream, inBaseDir, inEncoding):
        self.includeStack.append(
            (self.cursor, self.fileId, self.baseDir,
             self.encoding, self.stream))
        self.cursor = 0
        self.fileId = inFileId
        self.stream = inStream
        self.baseDir = inBaseDir
        self.encoding = inEncoding

    def popStream(self):
        if not self.includeStack:
            return False
        (self.cursor, self.fileId, self.baseDir,
         self.encoding, self.stream) = self.includeStack.pop()
        return True


class StreamReader:
    """This class handles the PSP source file.

    It provides the characters to the other parts of the system.
    It can move forward and backwards in a file and remember locations.
    """

    def __init__(self, filename, ctxt):
        self._pspFile = filename
        self._ctxt = ctxt
        self.sourceFiles = []
        self.current = None
        self.master = None

    def init(self):
        self.pushFile(self._ctxt.getFullPspFileName())

    def registerSourceFile(self, filepath):
        self.sourceFiles.append(filepath)
        return len(self.sourceFiles) - 1

    def pushFile(self, filepath, encoding=None):
        if self.master is None:
            parent = None
            self.master = filepath
        else:
            parent = os.path.split(self.master)[0]
        isAbsolute = os.path.isabs(filepath)
        if parent is not None and not isAbsolute:
            filepath = os.path.join(parent, filepath)
        fileId = self.registerSourceFile(filepath)
        with open(filepath, 'r') as handle:
            stream = handle.read()
            handle.seek(0, 0)
            if self.current is None:
                self.current = Mark(
                    self, fileId, stream, self._ctxt.getBaseUri(), encoding)
            else:
                self.current.pushStream(  # don't use yet
                    fileId, stream, self._ctxt.getBaseUri(), encoding)

    def popFile(self):
        if self.current is None:
            return None
        return self.current.popStream()

    def getFile(self, i):
        return self.sourceFiles[i]

    def newSourceFile(self, filename):
        if filename in self.sourceFiles:
            return None
        self.sourceFiles.append(filename)
        return len(self.sourceFiles)

    def mark(self):
        return copy(self.current)

    def skipUntil(self, s):
        """Greedy search.

        Return the point before the string, but move reader past it.
        """
        newCursor = self.current.stream.find(s, self.current.cursor)
        if newCursor < 0:
            self.current.cursor = len(self.current.stream)
            if self.hasMoreInput():
                self.popFile()
                self.skipUntil(s)
            else:
                raise EOFError
        else:
            self.current.cursor = newCursor
            mark = self.mark()
            self.current.cursor += len(s)
            return mark

    def reset(self, mark):
        self.current = mark

    def matches(self, s):
        if s == self.current.stream[
                self.current.cursor:self.current.cursor + len(s)]:
            return True
        return False

    def advance(self, length):
        """Advance length characters"""
        if length + self.current.cursor <= len(self.current.stream):
            self.current.cursor += length
        else:
            offset = len(self.current.stream) - self.current.cursor
            self.current.cursor = len(self.current.stream)
            if self.hasMoreInput():
                self.advance(length - offset)
            else:
                raise EOFError()

    def nextChar(self):
        if not self.hasMoreInput():
            return -1
        char = self.current.stream[self.current.cursor]
        self.advance(1)
        return char

    def isSpace(self):
        """No advancing."""
        return self.current.stream[self.current.cursor] in (' ', '\n')

    def isDelimiter(self):
        if not self.isSpace():
            c = self.peekChar()
            # Look for single character work delimiter:
            if c in ('=', '"', "'", '/'):
                return True
            # Look for end of comment or basic end tag:
            if c == '-':
                mark = self.mark()
                c = self.nextChar()
                try:
                    return c == '>' or (c == '-' and self.nextChar() == '>')
                finally:
                    self.reset(mark)
        else:
            return True

    def peekChar(self, cnt=1):
        if self.hasMoreInput():
            return self.current.stream[
                self.current.cursor:self.current.cursor+cnt]
        raise EOFError

    def skipSpaces(self):
        i = 0
        while self.isSpace():
            self.nextChar()
            i += 1
        return i

    def getChars(self, start, stop):
        mark = self.mark()
        self.reset(start)
        chars = self.current.stream[start.cursor:stop.cursor]
        self.reset(mark)
        return chars

    def hasMoreInput(self):
        if self.current.cursor >= len(self.current.stream):
            while self.popFile():
                if self.current.cursor < len(self.current.stream):
                    return True
            return False
        return True

    def nextContent(self):
        """Find next < char."""
        currentCursor = self.current.cursor
        self.current.cursor += 1
        newCursor = self.current.stream.find('<', self.current.cursor)
        if newCursor < 0:
            newCursor = len(self.current.stream)
        self.current.cursor = newCursor
        return self.current.stream[currentCursor:newCursor]

    def parseTagAttributes(self):
        """Parse the attributes at the beginning of a tag."""
        values = {}
        while True:
            self.skipSpaces()
            c = self.peekChar()
            if c == '>':
                return values
            if c == '-':
                mark = self.mark()
                self.nextChar()
                try:
                    if self.nextChar() == '-' and self.nextChar() == '>':
                        return values
                finally:
                    self.reset(mark)
            elif c == '%':
                mark = self.mark()
                self.nextChar()
                try:
                    if self.peekChar() == '>':
                        return values
                finally:
                    self.reset(mark)
            elif not c:
                break
            self.parseAttributeValue(values)
        raise PSPParserException('Unterminated attribute')

    def parseAttributeValue(self, valueDict):
        self.skipSpaces()
        name = self.parseToken(0)
        self.skipSpaces()
        if self.peekChar() != '=':
            raise PSPParserException('No attribute value')
        self.nextChar()
        self.skipSpaces()
        value = self.parseToken(1)
        self.skipSpaces()
        valueDict[name] = value

    def parseToken(self, quoted):
        # This may not be quite right:
        buffer = []
        self.skipSpaces()
        c = self.peekChar()
        if quoted:
            if c in ('"', "'"):
                endQuote = c
                self.nextChar()
                c = self.peekChar()
                while c is not None and c != endQuote:
                    c = self.nextChar()
                    if c == '\\':
                        c = self.nextChar()
                    buffer.append(c)
                    c = self.peekChar()
                if c is None:
                    raise PSPParserException('Unterminated attribute value')
                self.nextChar()
        else:
            if not self.isDelimiter():
                while not self.isDelimiter():
                    c = self.nextChar()
                    if c == '\\':
                        c = self.peekChar()
                        if c in ('"', "'", '>', '%'):
                            c = self.nextChar()
                    buffer.append(c)
        return ''.join(buffer)



import setuptools

with open('webware/Properties.py') as f:
    properties = {}
    exec(f.read(), properties)
    version = properties['version']
    version = '.'.join(map(str, version[:3])) + '.'.join(version[3:])
    description = properties['synopsis']

with open('README.md') as fh:
    long_description = fh.read()

requireDev = [
    'Pygments>=2.7,<3', 'WebTest>=2.0,<3',
    'waitress>=1.4.4,<2', 'hupper>=1.10,<2',
]
requireDocs = [
    'Sphinx>=3,<4', 'sphinx_rtd_theme>=0.5'
]
requireExamples = [
    'DBUtils>=2,<4', 'dominate>=2.5,<3', 'yattag>=1.14,<2',
    'Pygments>=2.7,<3', 'Pillow>=8,<9'
]
requireTests = [
    'psutil>=5.7,<6', 'flake8>=3.8,<4', 'pylint>=2.6,<3', 'tox>=3.20,<4',
    'pywin32>=228,<300;'
    'sys_platform=="win32" and implementation_name=="cpython"'
] + requireDev + requireDocs + requireExamples


setuptools.setup(
    name='Webware-for-Python',
    version=version,
    author='Christoph Zwerschke et al.',
    author_email='cito@online.de',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='web framework servlets',
    url='https://webwareforpython.github.io/w4py3/',
    project_urls={
        'Source': 'https://github.com/WebwareForPython/w4py3',
        'Issues': 'https://github.com/WebwareForPython/w4py3/issues',
        'Documentation': 'https://webwareforpython.github.io/w4py3/',
    },
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
    ],
    extras_require={
        'dev': requireDev,
        'docs': requireDocs,
        'examples': requireExamples,
        'tests': requireTests,
    },
    entry_points={
        'console_scripts': [
            'webware = webware.Scripts.WebwareCLI:main'
        ],
        'webware.plugins': [
            'MiscUtils = webware.MiscUtils',
            'PSP = webware.PSP',
            'TaskKit = webware.TaskKit',
            'UserKit = webware.UserKit',
            'WebUtils = webware.WebUtils'
        ]
    }
)



"""HTTP requests"""

import cgi
import os
import sys
import traceback
from operator import itemgetter
from time import time
from http.cookies import SimpleCookie as Cookie

import HTTPResponse

from MiscUtils import NoDefault
from WebUtils import FieldStorage
from Request import Request

debug = False


class HTTPRequest(Request):
    """The base class for HTTP requests."""

    # region Initialization

    def __init__(self, requestDict=None):
        Request.__init__(self)
        self._stack = []
        if requestDict:
            # Dictionaries come in from interfaces like WSGI
            if requestDict['format'] != 'CGI':
                raise ValueError(
                    'Request dict for HTTPRequest must have CGI format')
            self._time = requestDict['time']
            self._environ = requestDict['environ']
            self._input = requestDict['input']
            self._requestID = requestDict['requestID']
            try:
                self._fields = FieldStorage.FieldStorage(
                    self._input, environ=self._environ,
                    keep_blank_values=True, strict_parsing=False)
            except Exception:
                # Protect the loading of fields with an exception handler,
                # because bad headers sometimes can break the field storage
                # (see also https://bugs.python.org/issue27777).
                self._fields = cgi.FieldStorage(keep_blank_values=True)
                traceback.print_exc(file=sys.stderr)
            self._cookies = Cookie()
            if 'HTTP_COOKIE' in self._environ:
                # Protect the loading of cookies with an exception handler,
                # because MSIE cookies sometimes can break the cookie module.
                try:
                    self._cookies.load(self._environ['HTTP_COOKIE'])
                except Exception:
                    traceback.print_exc(file=sys.stderr)
        else:
            # If there's no dictionary, we pretend we're a CGI script
            # and see what happens...
            self._time = time()
            self._environ = os.environ.copy()
            self._input = None
            self._fields = cgi.FieldStorage(keep_blank_values=True)
            self._cookies = Cookie()

        env = self._environ

        # Debugging
        if debug:
            with open('env.text', 'a') as f:
                f.write('>> env for request:\n')
                for key in sorted(env):
                    f.write(f'{key}: {env[key]!r}\n')
                f.write('\n')

        # Get servlet path and query string
        self._servletPath = env.get('SCRIPT_NAME', '')
        self._pathInfo = env.get('PATH_INFO', '')
        self._extraURLPath = ''  # will be determined later
        self._queryString = env.get('QUERY_STRING', '')
        if 'REQUEST_URI' in env:
            self._uri = env['REQUEST_URI']
            # correct servletPath if there was a redirection
            if not (self._uri + '/').startswith(self._servletPath + '/'):
                i = self._uri.find(self._pathInfo)
                self._servletPath = self._uri[:i] if i > 0 else ''
        else:
            # REQUEST_URI isn't actually part of the CGI standard and some
            # web servers like IIS don't set it (as of 8/22/2000).
            if 'SCRIPT_URL' in env:
                self._uri = env['SCRIPT_URL']
                # correct servletPath if there was a redirection
                if not (self._uri + '/').startswith(self._servletPath + '/'):
                    i = self._uri.find(self._pathInfo)
                    self._servletPath = self._uri[:i] if i > 0 else ''
            else:
                self._uri = self._servletPath + self._pathInfo
            if self._queryString:
                self._uri += '?' + self._queryString

        # We use the cgi module to get the fields,
        # but then change them into an ordinary dictionary of values:
        fieldStorage, fields = self._fields, {}
        try:
            # Avoid accessing fieldStorage as dict; that would be very slow
            # as it always iterates over all items to find a certain key.
            # Instead, iterate directly over the items of the internal list.
            fieldItems = fieldStorage.list
        except AttributeError:
            # This can happen if we do not have a a regular POST
            # from an HTML form, but, for example, an XML-RPC request.
            fieldItems = None
            if debug:
                print("Cannot get fieldStorage list.")
        if fieldItems:
            for item in fieldItems:
                if item.filename:
                    if debug:
                        print("Uploaded file found:", item.filename)
                    fields.setdefault(item.name, []).append(item)
                else:
                    fields.setdefault(item.name, []).append(item.value)
            for key, value in fields.items():
                if len(value) == 1:
                    fields[key] = value[0]
        self._fieldStorage, self._fields = fieldStorage, fields

        # We use Tim O'Malley's Cookie class to get the cookies,
        # but then change them into an ordinary dictionary of values
        self._cookies = dict(
            (key, self._cookies[key].value) for key in self._cookies)

        self._contextName = None
        self._serverSidePath = self._serverSideContextPath = None
        self._serverRootPath = ''
        self._sessionExpired = False

        self._pathInfo = self.pathInfo()

        if debug:
            print("Done setting up request, found fields:", ', '.join(fields))

    # endregion Initialization

    # region Protocol

    def protocol(self):
        """Return the name and version of the protocol."""
        return self._environ.get('SERVER_PROTOCOL', 'HTTP/1.0')

    # endregion Protocol

    # region Security

    def isSecure(self):
        """Check whether this is a HTTPS connection."""
        return self._environ.get('HTTPS', '').lower() == 'on'

    # endregion Security

    # region Transactions

    def responseClass(self):
        return HTTPResponse.HTTPResponse

    # endregion Transactions

    # region Values

    def value(self, name, default=NoDefault):
        """Return the value with the given name.

        Values are fields or cookies.
        Use this method when you're field/cookie agnostic.
        """
        if name in self._fields:
            return self._fields[name]
        return self.cookie(name, default)

    def hasValue(self, name):
        """Check whether there is a value with the given name."""
        return name in self._fields or name in self._cookies

    def extraURLPath(self):
        """Return additional path components in the URL.

        Only works if the Application.config setting "ExtraPathInfo"
        is set to true; otherwise you will get a page not found error.
        """
        return self._extraURLPath

    # endregion Values

    # region Fields

    def fieldStorage(self):
        return self._fieldStorage

    def field(self, name, default=NoDefault):
        if default is NoDefault:
            return self._fields[name]
        return self._fields.get(name, default)

    def hasField(self, name):
        return name in self._fields

    def fields(self):
        return self._fields

    def setField(self, name, value):
        self._fields[name] = value

    def delField(self, name):
        del self._fields[name]

    # endregion Fields

    # region Cookies

    def cookie(self, name, default=NoDefault):
        """Return the value of the specified cookie."""
        if default is NoDefault:
            return self._cookies[name]
        return self._cookies.get(name, default)

    def hasCookie(self, name):
        """Return whether a cookie with the given name exists."""
        return name in self._cookies

    def cookies(self):
        """Return a dict of all cookies the client sent with this request."""
        return self._cookies

    # endregion Cookies

    # region Variables passed by server

    def serverDictionary(self):
        """Return a dictionary with the data the web server gave us.

        This data includes HTTP_HOST and HTTP_USER_AGENT, for example.
        """
        return self._environ

    # endregion Variables passed by server

    # region Sessions

    def session(self):
        """Return the session associated with this request.

        The session is either as specified by sessionId() or newly created.
        This is a convenience for transaction.session()
        """
        return self._transaction.session()

    def isSessionExpired(self):
        """Return whether the request originally had an expired session ID.

        Only works if the Application.config setting "IgnoreInvalidSession"
        is set to true; otherwise you get a canned error page on an invalid
        session, so your servlet never gets processed.
        """
        return self._sessionExpired

    def setSessionExpired(self, sessionExpired):
        self._sessionExpired = sessionExpired

    # endregion Sessions

    # region  Remote info

    def remoteUser(self):
        """Always returns None since authentication is not yet supported.

        Take from CGI variable REMOTE_USER.
        """
        return self._environ['REMOTE_USER']

    def remoteAddress(self):
        """Return a string containing the IP address of the client."""
        return self._environ['REMOTE_ADDR']

    def remoteName(self):
        """Return the fully qualified name of the client that sent the request.

        Returns the IP address of the client if the name cannot be determined.
        """
        env = self._environ
        return env.get('REMOTE_NAME', env['REMOTE_ADDR'])

    def accept(self, which=None):
        """Return preferences as requested by the user agent.

        The accepted preferences are returned as a list of codes
        in the same order as they appeared in the header.
        In other words, the explicit weighting criteria are ignored.

        If you do not define otherwise which preferences you are
        interested in ('language', 'charset', 'encoding'), by default
        you will get the user preferences for the content types.
        """
        var = 'HTTP_ACCEPT'
        if which:
            var += '_' + which.upper()
        prefs = []
        for pref in self._environ.get(var, '').split(','):
            pref = pref.partition(';')[0].strip()
            prefs.append(pref)
        return prefs

    # endregion  Remote info

    # region Path

    def urlPath(self):
        """Get the URL path relative to the mount point, without query string.

        This is actually the same as pathInfo().

        For example, http://host/Webware/Context/Servlet?x=1
        yields '/Context/Servlet'.
        """
        return self._pathInfo

    def urlPathDir(self):
        """Same as urlPath, but only gives the directory.

        For example, http://host/Webware/Context/Servlet?x=1
        yields '/Context'.
        """
        return os.path.dirname(self.urlPath())

    def setURLPath(self, path):
        """Set the URL path of the request.

        There is rarely a need to do this. Proceed with caution.
        """
        self._pathInfo = path
        self._uri = self._servletPath + path
        if self._queryString:
            self._uri += '?' + self._queryString

    def serverSidePath(self, path=None):
        """Return the absolute server-side path of the request.

        If the optional path is passed in, then it is joined with the
        server side directory to form a path relative to the object.
        """
        if path:
            if path.startswith('/'):
                path = path[1:]
            return os.path.normpath(os.path.join(
                os.path.dirname(self._serverSidePath), path))
        return self._serverSidePath

    def serverSideContextPath(self, path=None):
        """Return the absolute server-side path of the context of this request.

        If the optional path is passed in, then it is joined with the server
        side context directory to form a path relative to the object.

        This directory could be different from the result of serverSidePath()
        if the request is in a subdirectory of the main context directory.
        """
        if path:
            if path.startswith('/'):
                path = path[1:]
            return os.path.normpath(os.path.join(
                self._serverSideContextPath, path))
        return self._serverSideContextPath

    def contextName(self):
        """Return the name of the context of this request.

        This isn't necessarily the same as the name of the directory
        containing the context.
        """
        return self._contextName

    def servletURI(self):
        """Return servlet URI without any query strings or extra path info."""
        p = self._pathInfo
        if not self._extraURLPath:
            if p.endswith('/'):
                p = p[:-1]
            return p
        i = p.rfind(self._extraURLPath)
        if i >= 0:
            p = p[:i]
        if p.endswith('/'):
            p = p[:-1]
        return p

    def uriWebwareRoot(self):
        """Return relative URL path of the Webware root location."""
        if not self._serverRootPath:
            self._serverRootPath = ''
            loc = self.urlPath()
            loc, curr = os.path.split(loc)
            while True:
                loc, curr = os.path.split(loc)
                if not curr:
                    break
                self._serverRootPath += "../"
        return self._serverRootPath

    def scheme(self):
        """Return the URI scheme of the request (http or https)."""
        return 'https' if self.isSecure() else 'http'

    def hostAndPort(self):
        """Return the hostname and port part from the URL of this request."""
        env = self._environ
        if 'HTTP_HOST' in env:
            return env['HTTP_HOST']
        if 'SERVER_NAME' in env:
            return env['SERVER_NAME']
        host = env.get('SERVER_ADDR', '') or 'localhost'
        port = env.get('SERVER_PORT', '')
        defaultPort = '443' if self.isSecure() else '80'
        if port and port != defaultPort:
            host += ':' + port
        return host

    def serverURL(self, canonical=False):
        """Return the full Internet path to this request.

        This is the URL that was actually received by the web server
        before any rewriting took place. If canonical is set to true,
        then the canonical hostname of the server is used if possible.

        The path is returned without any extra path info or query strings,
        i.e. http://www.my.own.host.com:8080/Webware/TestPage.py
        """
        if canonical and 'SCRIPT_URI' in self._environ:
            return self._environ['SCRIPT_URI']
        return f'{self.scheme()}://{self.hostAndPort()}{self.serverPath()}'

    def serverURLDir(self):
        """Return the directory of the URL in full Internet form.

        Same as serverURL, but removes the actual page.
        """
        fullURL = self.serverURL()
        if fullURL and not fullURL.endswith('/'):
            fullURL = fullURL[:fullURL.rfind('/') + 1]
        return fullURL

    def serverPath(self):
        """Return the web server URL path of this request.

        This is the URL that was actually received by the web server
        before any rewriting took place.

        Same as serverURL, but without scheme and host.
        """
        if 'SCRIPT_URL' in self._environ:
            return self._environ['SCRIPT_URL']
        return self._servletPath + (
            self._stack[0][1] if self._stack else self._pathInfo)

    def serverPathDir(self):
        """Return the directory of the web server URL path.

        Same as serverPath, but removes the actual page.
        """
        fullURL = self.serverPath()
        if fullURL and not fullURL.endswith('/'):
            fullURL = fullURL[:fullURL.rfind('/') + 1]
        return fullURL

    def siteRoot(self):
        """Return the relative URL path of the home location.

        This includes all URL path components necessary to get back home
        from the current location.

        Examples:
            ''
            '../'
            '../../'

        You can use this as a prefix to a URL that you know is based off
        the home location. Any time you are in a servlet that may have been
        forwarded to from another servlet at a different level, you should
        prefix your URL's with this. That is, if servlet "Foo/Bar" forwards
        to "Qux", then the qux servlet should use siteRoot() to construct all
        links to avoid broken links. This works properly because this method
        computes the path based on the _original_ servlet, not the location
        of the servlet that you have forwarded to.
        """
        url = self.originalURLPath()
        if url.startswith('/'):
            url = url[1:]
        contextName = self.contextName() + '/'
        if url.startswith(contextName):
            url = url[len(contextName):]
        numStepsBack = url.count('/')
        return '../' * numStepsBack

    def siteRootFromCurrentServlet(self):
        """Return relative URL path to home seen from the current servlet.

        This includes all URL path components necessary to get back home
        from the current servlet (not from the original request).

        Similar to siteRoot() but instead, it returns the site root
        relative to the _current_ servlet, not the _original_ servlet.
        """
        url = self.urlPath()
        if url.startswith('/'):
            url = url[1:]
        contextName = self.contextName() + '/'
        if url.startswith(contextName):
            url = url[len(contextName):]
        numStepsBackward = url.count('/')
        return '../' * numStepsBackward

    def servletPathFromSiteRoot(self):
        """Return the "servlet path" of this servlet relative to the siteRoot.

        In other words, everything after the name of the context (if present).
        If you append this to the result of self.siteRoot() you get back to
        the current servlet. This is useful for saving the path to the current
        servlet in a database, for example.
        """
        urlPath = self.urlPath()
        if urlPath.startswith('/'):
            urlPath = urlPath[1:]
        parts = urlPath.split('/')
        newParts = []
        for part in parts:
            if part == '..' and newParts:
                newParts.pop()
            elif part != '.':
                newParts.append(part)
        if newParts[:1] == [self.contextName()]:
            newParts[:1] = []
        return '/'.join(newParts)

    # endregion Path

    # region Special

    def scriptName(self):
        """Return the name of the WSGI script as it appears in the URL.

        Example: '/Webware'
        Does not reflect redirection by the web server.
        Equivalent to the CGI variable SCRIPT_NAME.
        """
        return self._environ.get('SCRIPT_NAME', '')

    def scriptFileName(self):
        """Return the filesystem path of the WSGI script.

        Equivalent to the CGI variable SCRIPT_FILENAME.
        """
        return self._environ.get('SCRIPT_FILENAME', '')

    def environ(self):
        """Get the environment for the request."""
        return self._environ

    def push(self, servlet, url=None):
        """Push servlet and URL path on a stack, setting a new URL."""
        self._stack.append((servlet, self.urlPath(), self._contextName,
                            self._serverSidePath, self._serverSideContextPath,
                            self._serverRootPath, self._extraURLPath))
        if url is not None:
            self.setURLPath(url)

    def pop(self):
        """Pop URL path and servlet from the stack, returning the servlet."""
        if self._stack:
            (servlet, url, self._contextName,
             self._serverSidePath, self._serverSideContextPath,
             self._serverRootPath, self._extraURLPath) = self._stack.pop()
            if url is not None:
                self.setURLPath(url)
            return servlet

    def servlet(self):
        """Get current servlet for this request."""
        return self._transaction.servlet()

    def originalServlet(self):
        """Get original servlet before any forwarding."""
        if self._stack:
            return self._stack[0][0]
        return self.servlet()

    def previousServlet(self):
        """Get the servlet that passed this request to us, if any."""
        if self._stack:
            return self._stack[-1][0]

    parent = previousServlet  # kept old name as synonym

    def previousServlets(self):
        """Get the list of all previous servlets."""
        return [s[0] for s in self._stack]

    parents = previousServlets  # kept old name as synonym

    def originalURLPath(self):
        """Get URL path of the original servlet before any forwarding."""
        if self._stack:
            return self._stack[0][1]
        return self.urlPath()

    def previousURLPath(self):
        """Get the previous URL path, if any."""
        if self._stack:
            return self._stack[-1][1]

    def previousURLPaths(self):
        """Get the list of all previous URL paths."""
        return [s[1] for s in self._stack]

    def originalURI(self):
        """Get URI of the original servlet before any forwarding."""
        if self._stack:
            return self._servletPath + self._stack[0][1]
        return self.uri()

    def previousURI(self):
        """Get the previous URI, if any."""
        if self._stack:
            return self._servletPath + self._stack[-1][1]

    def previousURIs(self):
        """Get the list of all previous URIs."""
        return [self._servletPath + s[1] for s in self._stack]

    def originalContextName(self):
        """Return the name of the original context before any forwarding."""
        if self._stack:
            return self._stack[0][2]
        return self._contextName

    def previousContextName(self):
        """Get the previous context name, if any."""
        if self._stack:
            return self._stack[-1][2]

    def previousContextNames(self):
        """Get the list of all previous context names."""
        return [s[2] for s in self._stack]

    def rawInput(self, rewind=False):
        """Get the raw input from the request.

        This gives you a file-like object for the data that was sent with
        the request (e.g., the body of a POST request, or the document
        uploaded in a PUT request).

        The file might not be rewound to the beginning if there was valid,
        form-encoded POST data. Pass rewind=True if you want to be sure
        you get the entire body of the request.
        """
        fs = self.fieldStorage()
        if fs is None:
            return None
        if rewind and fs.file:
            fs.file.seek(0)
        return fs.file

    def time(self):
        """Return the time that the request was received."""
        return self._time

    def requestID(self):
        """Return the request ID.

        The request ID is a serial number unique to this request
        (at least unique for one run of the Application).
        """
        return self._requestID

    # endregion Special

    # region Information

    def servletPath(self):
        """Return the base URL for the servlets, sans host.

        This is useful in cases when you are constructing URLs.
        See Testing/Main.py for an example use.

        Roughly equivalent to the CGI variable SCRIPT_NAME,
        but reflects redirection by the web server.
        """
        return self._servletPath

    def contextPath(self):
        """Return the portion of the URI that is the context of the request."""
        return self._serverSideContextPath

    def pathInfo(self):
        """Return any extra path information as sent by the client.

        This is anything after the servlet name but before the query string.
        Equivalent to the CGI variable PATH_INFO.
        """
        return self._pathInfo

    def pathTranslated(self):
        """Return extra path information translated as file system path.

        This is the same as pathInfo() but translated to the file system.
        Equivalent to the CGI variable PATH_TRANSLATED.
        """
        return self._environ.get('PATH_TRANSLATED', '')

    def queryString(self):
        """Return the query string portion of the URL for this request.

        Equivalent to the CGI variable QUERY_STRING.
        """
        return self._queryString

    def uri(self):
        """Return the URI for this request (everything after the host name).

        This is the URL that was actually received by the web server
        before any rewriting took place, including the query string.
        Equivalent to the CGI variable REQUEST_URI.
        """
        return self._uri

    def method(self):
        """Return the HTTP request method (in all uppercase).

        Typically from the set GET, POST, PUT, DELETE, OPTIONS and TRACE.
        """
        return self._environ['REQUEST_METHOD'].upper()

    def sessionId(self):
        """Return a string with the session ID specified by the client.

        Returns None if there is no session ID.
        """
        trans = self._transaction
        app = trans.application()
        sid = self.value(app.sessionName(trans), None)
        if app.setting('Debug')['Sessions']:
            print('>> sessionId: returning sid =', sid)
        return sid

    def setSessionId(self, sessionID, force=False):
        """Set the session ID.

        This needs to be called _before_ attempting to use the session.
        This would be useful if the session ID is being passed in through
        unusual means, for example via a field in an XML-RPC request.

        Pass in force=True if you want to force this session ID to be used
        even if the session doesn't exist. This would be useful in unusual
        circumstances where the client is responsible for creating the unique
        session ID rather than the server.
        Be sure to use only legal filename characters in the session ID --
        0-9, a-z, A-Z, _, -, and . are OK but everything else will be rejected,
        as will identifiers longer than 80 characters.
        (Without passing in force=True, a random session ID will be generated
        if that session ID isn't already present in the session store.)
        """
        # Modify the request so that it looks like a hashed version of the
        # given session ID was passed in
        trans = self._transaction
        app = trans.application()
        self.setField(app.sessionName(trans), sessionID)
        if force:
            # Force this session ID to exist, so that a random session ID
            # won't be created in case it's a new session.
            app.createSessionWithID(trans, sessionID)

    # endregion Information

    # region Inspection

    def info(self):
        """Return request info.

        Return a list of tuples where each tuple has a key/label (a string)
        and a value (any kind of object).

        Values are typically atomic values such as numbers and strings or
        another list of tuples in the same fashion. This is for debugging only.
        """
        info = [
            ('time', self._time),
            ('environ', self._environ),
            ('input', self._input),
            ('fields', self._fields),
            ('cookies', self._cookies)
        ]
        for method in _infoMethods:
            try:
                info.append((method.__name__, method(self)))
            except Exception:
                info.append((method.__name__, None))
        return info

    def htmlInfo(self):
        """Return a single HTML string that represents info().

        Useful for inspecting objects via web browsers.
        """
        return htmlInfo(self.info())

    _exceptionReportAttrNames = Request._exceptionReportAttrNames + [
        'uri', 'scriptName', 'servletPath', 'serverSidePath'
        'pathInfo', 'pathTranslated', 'queryString', 'method',
        'sessionId', 'previousURLPaths', 'fields', 'cookies', 'environ']

    # endregion Inspection


# region Info structure

_infoMethods = (
    HTTPRequest.scriptName,
    HTTPRequest.servletPath,
    HTTPRequest.contextPath,
    HTTPRequest.pathInfo,
    HTTPRequest.pathTranslated,
    HTTPRequest.queryString,
    HTTPRequest.uri,
    HTTPRequest.method,
    HTTPRequest.sessionId
)


def htmlInfo(info):
    """Return a single HTML string that represents the info structure.

    Useful for inspecting objects via web browsers.
    """
    res = ['<table>\n']
    for pair in info:
        value = pair[1]
        if hasattr(value, 'items') and (
                isinstance(value, dict) or hasattr(value, '__getitem__')):
            value = htmlInfo(_infoForDict(value))
        res.append(
            f'<tr style="vertical-align:top"><td>{pair[0]}</td>'
            f'<td>{value}&nbsp;</td></tr>\n')
    res.append('</table>\n')
    return ''.join(res)


def _infoForDict(d):
    """Return an "info" structure for any dictionary-like object."""
    return sorted(d.items, key=itemgetter(0))

# endregion Info structure



"""Implementation of client sessions."""

import re
from time import time, localtime
from urllib import parse

from MiscUtils import NoDefault
from MiscUtils.Funcs import uniqueId


class SessionError(Exception):
    """Client session error"""


class Session:
    """Implementation of client sessions.

    All methods that deal with time stamps, such as creationTime(),
    treat time as the number of seconds since January 1, 1970.

    Session identifiers are stored in cookies. Therefore, clients
    must have cookies enabled.

    Note that the session id should be a string that is valid
    as part of a filename. This is currently true, and should
    be maintained if the session id generation technique is
    modified. Session ids can be used in filenames.
    """

    # region Init

    def __init__(self, trans, identifier=None):
        self._lastAccessTime = self._creationTime = time()
        self._isExpired = self._dirty = False
        self._numTrans = 0
        self._values = {}
        app = trans.application()
        self._timeout = app.sessionTimeout(trans)
        self._prefix = app.sessionPrefix(trans)
        self._sessionName = app.sessionName(trans)

        if identifier:
            if re.search(r'[^\w\.\-]', identifier) is not None:
                raise SessionError("Illegal characters in session identifier")
            if len(identifier) > 80:
                raise SessionError("Session identifier too long")
            self._identifier = identifier
        else:
            attempts = 0
            while attempts < 10000:
                self._identifier = self._prefix + (
                    '{:02d}{:02d}{:02d}{:02d}{:02d}{:02d}').format(
                        *localtime()[:6]) + '-' + uniqueId(self)
                if not app.hasSession(self._identifier):
                    break
                attempts += 1
            else:
                raise SessionError(
                    "Can't create valid session id"
                    f" after {attempts} attempts.")

        if app.setting('Debug')['Sessions']:
            print('>> [session] Created session, timeout =', self._timeout,
                  'id =', self._identifier, 'self =', self)

    # endregion Init

    # region Access

    def creationTime(self):
        """Return the time when this session was created."""
        return self._creationTime

    def lastAccessTime(self):
        """Get last access time.

        Returns the last time the user accessed the session through
        interaction. This attribute is updated in awake(), which is
        invoked at the beginning of a transaction.
        """
        return self._lastAccessTime

    def identifier(self):
        """Return a string that uniquely identifies the session.

        This method will create the identifier if needed.
        """
        return self._identifier

    def isDirty(self):
        """Check whether the session is dirty (has unsaved changes)."""
        return self._dirty

    def setDirty(self, dirty=True):
        """Set the dirty status of the session."""
        self._dirty = dirty

    def isExpired(self):
        """Check whether the session has been previously expired.

        See also: expiring()
        """
        return getattr(self, '_isExpired', False) or self._timeout == 0

    def isNew(self):
        """Check whether the session is new."""
        return self._numTrans < 2

    def timeout(self):
        """Return the timeout for this session in seconds."""
        return self._timeout

    def setTimeout(self, timeout):
        """Set the timeout on this session in seconds."""
        self._timeout = timeout

    # endregion Access

    # region Invalidate

    def invalidate(self):
        """Invalidate the session.

        It will be discarded the next time it is accessed.
        """
        self._lastAccessTime = 0
        self._values = {}
        self._dirty = False
        self._timeout = 0

    # endregion Invalidate

    # region Values

    def value(self, name, default=NoDefault):
        if default is NoDefault:
            return self._values[name]
        return self._values.get(name, default)

    def hasValue(self, name):
        return name in self._values

    def setValue(self, name, value):
        self._values[name] = value
        self._dirty = True

    def delValue(self, name):
        del self._values[name]
        self._dirty = True

    def values(self):
        return self._values

    def __getitem__(self, name):
        return self.value(name)

    def __setitem__(self, name, value):
        self.setValue(name, value)

    def __delitem__(self, name):
        self.delValue(name)

    def __contains__(self, name):
        return self.hasValue(name)

    # endregion Values

    # region Transactions

    def awake(self, _trans):
        """Let the session awake.

        Invoked during the beginning of a transaction, giving a Session an
        opportunity to perform any required setup. The default implementation
        updates the 'lastAccessTime'.
        """
        self._lastAccessTime = time()
        self._numTrans += 1

    def respond(self, trans):
        """Let the session respond to a request.

        The default implementation does nothing, but could do something
        in the future. Subclasses should invoke super.
        """
        # base method does nothing

    def sleep(self, trans):
        """Let the session sleep again.

        Invoked during the ending of a transaction, giving a Session an
        opportunity to perform any required shutdown. The default
        implementation does nothing, but could do something in the future.
        Subclasses should invoke super.
        """
        # base method does nothing

    def expiring(self):
        """Let the session expire.

        Called when session is expired by the application.
        Subclasses should invoke super.
        Session store __delitem__()s should invoke if not isExpired().
        """
        self._isExpired = True

    def numTransactions(self):
        """Get number of transactions.

        Returns the number of transactions in which the session has been used.
        """
        return self._numTrans

    # endregion Transactions

    # region Utility

    def sessionEncode(self, url):
        """Encode the session ID as a parameter to a url."""
        url = list(parse.urlparse(url))  # make a list
        if url[4]:
            url[4] += '&'
        url[4] += f'{self._sessionName}={self.identifier()}'
        url = parse.urlunparse(url)
        return url

    # endregion Utility

    # region Exception reports

    _exceptionReportAttrNames = [
        'isDirty', 'isExpired', 'lastAccessTime',
        'numTransactions', 'timeout', 'values']

    def writeExceptionReport(self, handler):
        handler.writeTitle(self.__class__.__name__)
        handler.writeAttrs(self, self._exceptionReportAttrNames)

    # endregion Exception reports



from URLParser import URLParser


class VHostParser(URLParser):

    def parseHook(self, trans, requestPath, app, hook):
        host = trans.request().environ()['HTTP_HOST']
        hook.parse(trans, f'/{host}{requestPath}', app)


urlParserHook = VHostParser()



"""Event handler for parsing PSP tokens.

This module is called when the Parser encounters psp tokens.
It creates a generator to handle the PSP token. When the PSP
source file is fully parsed, this module calls all of the
generators in turn to output their source code.

Copyright (c) by Jay Love, 2000 (mailto:jsliv@jslove.org)

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee or royalty is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation or portions thereof, including modifications,
that you make.

This software is based in part on work done by the Jakarta group.
"""

import time

from .Generators import (
    ResponseObject, CharDataGenerator, EndBlockGenerator, ExpressionGenerator,
    IncludeGenerator, InsertGenerator, MethodGenerator, MethodEndGenerator,
    ScriptGenerator, ScriptFileGenerator, ScriptClassGenerator)


# pylint: disable=unused-argument

class ParseEventHandler:
    """This is a key class.

    It implements the handling of all the parsing elements.
    Note: This files JSP cousin is called ParseEventListener,
    I don't know why, but Handler seemed more appropriate to me.
    """

    aspace = ' '
    defaults = dict(
        BASE_CLASS='Page',
        BASE_METHOD='writeHTML',
        imports=dict(filename='classes'),
        threadSafe='no',
        instanceSafe='yes',
        indent=4,
        gobbleWhitespace=True,
        formatter='str'
    )

    def __init__(self, ctxt, parser):
        self._ctxt = ctxt
        self._gens = []
        self._reader = ctxt.getReader()
        self._writer = ctxt.getServletWriter()
        self._parser = parser
        self._imports = []
        self._importedSymbols = []
        self._baseMethod = self.defaults['BASE_METHOD']
        self._baseClasses = [self.defaults['BASE_CLASS']]
        self._threadSafe = self.defaults['threadSafe']
        self._instanceSafe = self.defaults['instanceSafe']
        self._indent = self.defaults['indent']
        self._gobbleWhitespace = self.defaults['gobbleWhitespace']
        self._formatter = self.defaults['formatter']

    def addGenerator(self, gen):
        self._gens.append(gen)

    def handleExpression(self, start, stop, attrs):
        """Flush any template data and create a new ExpressionGenerator."""
        self._parser.flushCharData(self.tmplStart, self.tmplStop)
        exp = ExpressionGenerator(self._reader.getChars(start, stop))
        self.addGenerator(exp)

    def handleCharData(self, start, stop, chars):
        """Flush character data into a CharDataGenerator."""
        if chars:
            gen = CharDataGenerator(chars)
            self.addGenerator(gen)

    def handleComment(self, start, stop):
        """Comments get swallowed into nothing."""
        self._parser.flushCharData(self.tmplStart, self.tmplStop)
        # do nothing, this just eats the comment

    def handleInclude(self, attrs, param):
        """This is for includes of the form <psp:include ...>

        This function essentially forwards the request to the specified
        URL and includes that output.
        """
        self._parser.flushCharData(self.tmplStart, self.tmplStop)
        gen = IncludeGenerator(attrs, param, self._ctxt)
        self.addGenerator(gen)

    def handleInsert(self, attrs, param):
        """This is for includes of the form <psp:insert ...>

        This type of include is not parsed, it is just inserted into the
        output stream.
        """
        self._parser.flushCharData(self.tmplStart, self.tmplStop)
        gen = InsertGenerator(attrs, param, self._ctxt)
        self.addGenerator(gen)

    def importHandler(self, imports, start, stop):
        importList = imports.split(',')
        for imp in importList:
            if ':' in imp:
                module, symbol = imp.split(':', 1)
                module, symbol = module.strip(), symbol.strip()
                self._importedSymbols.append(symbol)
                self._imports.append(f'from {module} import {symbol}')
            else:
                symbol = imp.strip()
                self._imports.append(f'import {symbol}')

    def extendsHandler(self, bases, start, stop):
        """Extends is a page directive.

        It sets the base class (or multiple base classes) for the class
        that this class will generate. The choice of base class affects
        the choice of a method to override with the BaseMethod page directive.
        The default base class is PSPPage. PSPPage inherits from Page.py.
        """
        self._baseClasses = [s.strip() for s in bases.split(',')]

    def mainMethodHandler(self, method, start, stop):
        """BaseMethod is a page directive.

        It sets the class method that the main body of this PSP page
        over-rides. The default is WriteHTML. This value should be set
        to either WriteHTML or writeBody. See the PSPPage.py and Page.py
        servlet classes for more information.
        """
        self._baseMethod = method

    def threadSafeHandler(self, value, start, stop):
        """Handle isThreadSage.

        isThreadSafe is a page directive.
        The value can be "yes" or "no".
        Default is no because the default base class,
        Page.py, isn't thread safe.
        """
        self._threadSafe = value

    def instanceSafeHandler(self, value, start, stop):
        """Handle isInstanceSafe.

        isInstanceSafe tells the Servlet engine whether it is safe
        to use object instances of this page multiple times.
        The default is "yes".

        Saying "no" here hurts performance.
        """
        self._instanceSafe = value

    def indentTypeHandler(self, indentType, start, stop):
        """Declare whether tabs are used to indent source code."""
        indentType = indentType.lower()
        if indentType not in ('tabs', 'spaces', 'braces'):
            raise TypeError('Invalid Indentation Type')
        self._writer.setIndentType(indentType)

    def indentSpacesHandler(self, amount, start, stop):
        """Set number of spaces used to indent in generated source."""
        self._indentSpaces = int(amount)  # don't really need this
        self._writer.setIndentSpaces(int(amount))

    def gobbleWhitespaceHandler(self, value, start, stop):
        """Declare whether whitespace between script tags are gobble up."""
        if value.lower() in ('no', 'false', '0'):
            self._gobbleWhitespace = False

    def formatterHandler(self, value, start, stop):
        """Set an alternate formatter function to use instead of str()."""
        self._formatter = value

    directiveHandlers = {
        'imports': importHandler,
        'import': importHandler,
        'extends': extendsHandler,
        'method': mainMethodHandler,
        'isThreadSafe': threadSafeHandler,
        'isInstanceSafe': instanceSafeHandler,
        'BaseClass': extendsHandler,
        'indentSpaces': indentSpacesHandler,
        'indentType': indentTypeHandler,
        'gobbleWhitespace': gobbleWhitespaceHandler,
        'formatter': formatterHandler}

    def handleDirective(self, directive, start, stop, attrs):
        """Flush any template data and create a new DirectiveGenerator."""
        self._parser.flushCharData(self.tmplStart, self.tmplStop)
        # big switch
        if directive == 'page':
            for h in attrs:
                if h in self.directiveHandlers:
                    self.directiveHandlers[h](self, attrs[h], start, stop)
                else:
                    raise ValueError(f'No page directive handler: {h}')
        elif directive == 'include':
            filename = attrs['file']
            encoding = attrs.get('encoding')
            try:
                self._reader.pushFile(filename, encoding)
            except IOError as e:
                raise IOError(f'PSP include file not found: {filename}') from e
        else:
            raise ValueError('Invalid directive: {directive}')

    def handleScript(self, start, stop, attrs):
        """Handle scripting elements"""
        self._parser.flushCharData(self.tmplStart, self.tmplStop)
        gen = ScriptGenerator(self._reader.getChars(start, stop), attrs)
        self.addGenerator(gen)

    def handleScriptFile(self, start, stop, attrs):
        """Python script that goes at the file/module level"""
        self._parser.flushCharData(self.tmplStart, self.tmplStop)
        gen = ScriptFileGenerator(self._reader.getChars(start, stop), attrs)
        self.addGenerator(gen)

    def handleScriptClass(self, start, stop, attrs):
        """Python script that goes at the class level"""
        self._parser.flushCharData(self.tmplStart, self.tmplStop)
        gen = ScriptClassGenerator(self._reader.getChars(start, stop), attrs)
        self.addGenerator(gen)

    def handleEndBlock(self):
        self._parser.flushCharData(self.tmplStart, self.tmplStop)
        gen = EndBlockGenerator()
        self.addGenerator(gen)

    def handleMethod(self, start, stop, attrs):
        self._parser.flushCharData(self.tmplStart, self.tmplStop)
        gen = MethodGenerator(self._reader.getChars(start, stop), attrs)
        self.addGenerator(gen)

    def handleMethodEnd(self, start, stop, attrs):
        # self._parser.flushCharData(self.tmplStart, self.tmplStop)
        gen = MethodEndGenerator(self._reader.getChars(start, stop), attrs)
        self.addGenerator(gen)

    # region generation of the page

    def beginProcessing(self):
        pass

    def endProcessing(self):
        encoding = self._ctxt.getPythonFileEncoding()
        if encoding:
            # add encoding as a hint for both Python (PEP263) and editors
            self._writer.println(f'# -*- coding: {encoding} -*-')
        ascTime = time.asctime(time.localtime(time.time()))
        self._writer.println(
            f'# Generated automatically by PSP compiler on {ascTime}')
        self._writer.println()
        self.generateHeader()
        self.generateAll('psp:file')
        self.generateDeclarations()  # overwrite when we can handle extends
        self.generateInitPSP()
        self.generateAll('psp:class')
        self.generateAll('Declarations')
        self._writer.println('\n')
        self.generateMainMethod()
        self.optimizeCharData()
        if self._gobbleWhitespace:
            self.gobbleWhitespace()
        self.generateAll('Service')
        self._writer.println()
        self.generateFooter()

    def setTemplateInfo(self, start, stop):
        """Mark non code data."""
        self.tmplStart = start
        self.tmplStop = stop

    def generateHeader(self):
        for imp in self._imports:
            self._writer.println(imp)
        self._writer.println('import Page')
        for baseClass in self._baseClasses:
            if '.' not in baseClass and baseClass not in self._importedSymbols:
                self._writer.println('import ' + baseClass)
        self._writer.println(
            f"__orig_file__ = {self._ctxt.getFullPspFileName()!r}")

    def generateDeclarations(self):
        # The PSP "extends" directive allows you to use a shortcut
        # -- if the module name is the same as the class name,
        # you can say "Classname" instead of "ClassName.ClassName".
        # But we can't tell right now which names are actually class names,
        # and which names are really module names that contain a class of
        # the same name. So we have to generate code that checks at runtime.
        writer = self._writer
        writer.println()
        writer.println('import types')
        writer.println('_baseClasses = []')
        for baseClass in self._baseClasses:
            className = baseClass.rpartition('.')[2]
            writer.println(
                f'if isinstance({baseClass}, types.ModuleType):')
            writer.pushIndent()
            writer.println(
                f'_baseClasses.append({baseClass}.{className})')
            writer.popIndent()
            writer.println('else:')
            writer.pushIndent()
            writer.println(f'_baseClasses.append({baseClass})')
            writer.popIndent()
        writer.println()
        # Now write the class line:
        writer.printChars('class ')
        writer.printChars(self._ctxt.getServletClassName())
        writer.printChars('(')
        writer.printChars(','.join(
            f'_baseClasses[{i:d}]' for i in range(len(self._baseClasses))))
        writer.printChars('):')
        writer.printChars('\n')
        writer.pushIndent()
        writer.println('def canBeThreaded(self):')  # sadly, still needed
        writer.pushIndent()
        if self._threadSafe.lower() in ('no', 'false', '0'):
            writer.println('return False')
        else:
            writer.println('return True')
        writer.popIndent()
        writer.println()
        writer.println('def canBeReused(self):')  # sadly, still needed
        writer.pushIndent()
        if self._instanceSafe.lower() in ('no', 'false', '0'):
            writer.println('return False')
        else:
            writer.println('return True')
        writer.popIndent()
        writer.println()
        if not writer._awakeCreated:
            writer.println('def awake(self, trans):')
            writer.pushIndent()
            writer.println('for baseclass in self.__class__.__bases__:')
            writer.pushIndent()
            writer.println('if hasattr(baseclass, "awake"):')
            writer.pushIndent()
            writer.println('baseclass.awake(self, trans)')
            writer.println('break\n')
            writer.popIndent()  # end if statement
            writer.popIndent()  # end for statement
            writer.println('self.initPSP()\n')
            writer.println()
            writer.popIndent()
            writer.println()
        writer.println('def __includeFile(self, filename):')
        writer.pushIndent()
        writer.println('self.write(open(filename).read())')
        writer.popIndent()
        writer.println()

    def generateInitPSP(self):
        writer = self._writer
        writer.println('def initPSP(self):\n')
        writer.pushIndent()
        writer.println('pass\n')  # nothing for now
        writer.popIndent()
        writer.println()

    def generateMainMethod(self):
        # Write the output method requested with <%@ page method=... %>
        writer = self._writer
        writer.printIndent()
        baseMethod = self._baseMethod
        writer.printChars(f'def {baseMethod}(self, transaction=None):\n')
        writer.pushIndent()
        writer.println('"""I take a Webware Transaction object."""')
        writer.println('trans = self._transaction')
        writer.println(ResponseObject + ' = trans.response()')
        writer.println('req = trans.request()')
        writer.println(f'self._{baseMethod}({ResponseObject}, req, trans)\n')
        # Put the real output code in a function that doesn't need
        # a 'transaction' for unit tests.
        writer.popIndent()
        writer.println(
            f'def _{baseMethod}(self, {ResponseObject},'
            ' req=None, trans=None):')
        writer.pushIndent()
        writer.println(
            '"""I take a file-like object. I am useful for unit testing."""')
        writer.println(f'_formatter = {self._formatter}')

    def generateFooter(self):
        self._writer.popIndent()
        self._writer.println('##footer')

    def generateAll(self, phase):
        for gen in self._gens:
            if gen.phase == phase:
                gen.generate(self._writer)

    def optimizeCharData(self):
        """Optimize the CharData.

        Too many char data generators make the servlet slow.
        If the current Generator and the next are both CharData type,
        merge their data.
        """
        gens = self._gens
        count = 0
        genCount = len(gens)
        while count < genCount-1:
            if (isinstance(gens[count], CharDataGenerator)
                    and isinstance(gens[count + 1], CharDataGenerator)):
                gens[count].mergeData(gens[count + 1])
                gens.remove(gens[count+1])
                genCount -= 1
            else:
                count += 1

    def gobbleWhitespace(self):
        """Gobble up whitespace.

        This method looks for a character block between two PSP blocks
        that contains only whitespace. If it finds one, it deletes it.

        This is necessary so that a write() line can't sneek in between
        a if/else, try/except etc.
        """
        debug = False
        gens = self._gens
        sideClasses = (ScriptGenerator, EndBlockGenerator)
        count = 1
        genCount = len(gens)
        if debug:
            for gen in gens:
                print(f"Generator type = {gen.__class__}")
        while count < genCount - 1:
            if (isinstance(gens[count], CharDataGenerator)
                    and gens[count + 1].__class__ in sideClasses
                    and gens[count - 1].__class__ in sideClasses):
                if checkForTextHavingOnlyGivenChars(gens[count].chars):
                    gens.remove(gens[count])
                    genCount -= 1
            count += 1

    # endregion generation of the page


def checkForTextHavingOnlyGivenChars(text, whitespace=None):
    """Checks whether text contains only whitespace (or other chars).

    Does the given text contain anything other than the whitespace characters?
    Return true if text is only whitespace characters.
    """
    if whitespace is None:
        return text.isspace()
    for char in text:
        if char not in whitespace:
            return False
    return True



"""The UserManagerToFile class."""

import os

from glob import glob
from pickle import load, dump

from MiscUtils import NoDefault
from MiscUtils.MixIn import MixIn

from .User import User
from .UserManager import UserManager


class UserManagerToFile(UserManager):
    """User manager storing user data in the file system.

    When using this user manager, make sure you invoke setUserDir()
    and that this directory is writeable by your application.
    It will contain one file per user with the user's serial number
    as the main filename and an extension of '.user'.

    The default user directory is the current working directory,
    but relying on the current directory is often a bad practice.
    """

    # region Init

    def __init__(self, userClass=None):
        UserManager.__init__(self, userClass=None)
        self.setEncoderDecoder(dump, load)
        self.setUserDir(os.getcwd())
        self.initNextSerialNum()

    def initNextSerialNum(self):
        if os.path.exists(self._userDir):
            serialNums = self.scanSerialNums()
            if serialNums:
                self._nextSerialNum = max(serialNums) + 1
            else:
                self._nextSerialNum = 1
        else:
            self._nextSerialNum = 1

    # endregion Init

    # region File storage specifics

    def userDir(self):
        return self._userDir

    def setUserDir(self, userDir):
        """Set the directory where user information is stored.

        You should strongly consider invoking initNextSerialNum() afterwards.
        """
        self._userDir = userDir

    def loadUser(self, serialNum, default=NoDefault):
        """Load the user with the given serial number from disk.

        If there is no such user, a KeyError will be raised unless
        a default value was passed, in which case that value is returned.
        """
        filename = f'{serialNum}.user'
        filename = os.path.join(self.userDir(), filename)
        if os.path.exists(filename):
            with open(filename, 'rb') as f:
                user = self.decoder()(f)
            self._cachedUsers.append(user)
            self._cachedUsersBySerialNum[serialNum] = user
            return user
        if default is NoDefault:
            raise KeyError(serialNum)
        return default

    def scanSerialNums(self):
        """Return a list of all the serial numbers of users found on disk.

        Serial numbers are always integers.
        """
        return [
            int(os.path.basename(num[:-5]))
            for num in glob(os.path.join(self.userDir(), '*.user'))]

    # endregion File storage specifics

    # region UserManager customizations

    def setUserClass(self, userClass):
        """Overridden to mix in UserMixIn to the class that is passed in."""
        MixIn(userClass, UserMixIn)
        UserManager.setUserClass(self, userClass)

    # endregion UserManager customizations

    # region UserManager concrete methods

    def nextSerialNum(self):
        result = self._nextSerialNum
        self._nextSerialNum += 1
        return result

    def addUser(self, user):
        if not isinstance(user, User):
            raise TypeError(f'{user} is not a User object')
        user.setSerialNum(self.nextSerialNum())
        user.externalId()  # set unique id
        UserManager.addUser(self, user)
        user.save()

    def userForSerialNum(self, serialNum, default=NoDefault):
        user = self._cachedUsersBySerialNum.get(serialNum)
        if user is not None:
            return user
        return self.loadUser(serialNum, default)

    def userForExternalId(self, externalId, default=NoDefault):
        for user in self._cachedUsers:
            if user.externalId() == externalId:
                return user
        for user in self.users():
            if user.externalId() == externalId:
                return user
        if default is NoDefault:
            raise KeyError(externalId)
        return default

    def userForName(self, name, default=NoDefault):
        for user in self._cachedUsers:
            if user.name() == name:
                return user
        for user in self.users():
            if user.name() == name:
                return user
        if default is NoDefault:
            raise KeyError(name)
        return default

    def users(self):
        return _UserList(self)

    def activeUsers(self):
        return _UserList(self, lambda user: user.isActive())

    def inactiveUsers(self):
        return _UserList(self, lambda user: not user.isActive())

    # endregion UserManager concrete methods

    # region Encoder/decoder

    def encoder(self):
        return self._encoder

    def decoder(self):
        return self._decoder

    def setEncoderDecoder(self, encoder, decoder):
        self._encoder = encoder
        self._decoder = decoder

    # endregion Encoder/decoder


class UserMixIn:

    def filename(self):
        return os.path.join(
            self.manager().userDir(), f'{self.serialNum()}.user')

    def save(self):
        with open(self.filename(), 'wb') as f:
            self.manager().encoder()(self, f)


class _UserList:

    def __init__(self, mgr, filterFunc=None):
        self._mgr = mgr
        self._serialNums = mgr.scanSerialNums()
        self._count = len(self._serialNums)
        self._data = None
        if filterFunc:
            results = []
            for user in self:
                if filterFunc(user):
                    results.append(user)
            self._count = len(results)
            self._data = results

    def __getitem__(self, index):
        if index >= self._count:
            raise IndexError(index)
        if self._data:
            # We have the data directly. Just return it.
            return self._data[index]
        # We have a list of the serial numbers.
        # Get the user from the manager via the cache or loading.
        serialNum = self._serialNums[index]
        if serialNum in self._mgr._cachedUsersBySerialNum:
            return self._mgr._cachedUsersBySerialNum[serialNum]
        return self._mgr.loadUser(self._serialNums[index])

    def __len__(self):
        return self._count



from functools import reduce
from operator import truediv

from webware.XMLRPCServlet import XMLRPCServlet


class XMLRPCExample(XMLRPCServlet):
    """Example XML-RPC servlet.

    To try it out, use something like the following:

    >>> from xmlrpc.client import ServerProxy as Server
    >>> server = Server('http://localhost:8080/Examples/XMLRPCExample')
    >>> server.multiply(10,20)
    200
    >>> server.add(10,20)
    30

    You'll get an exception if you try to call divide, because that
    method is not listed in exposedMethods.
    """

    def exposedMethods(self):
        return ['multiply', 'add']

    def multiply(self, x, y):
        return x * y

    def add(self, x, y):
        return x + y

    def divide(self, *args):
        return reduce(truediv, args)
